#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from aiohomekit.crypto.srp import SrpClient, SrpServer

# To find short keys
# for _ in range(500000):
#    srp = SrpClient("Pair-Setup", "123-45-789")
#    pub_key_bytes = SrpClient.to_byte_array(srp.A)
#    if len(pub_key_bytes) < 384:
#        pprint.pprint(["found key", srp.a])


class ZeroSaltSrpServer(SrpServer):
    def _create_salt_bytes(self):
        return b"\x00" * 16


class LeadingZeroPrivateKeySrpClient(SrpClient):
    def generate_private_key(self):
        return 292137137271783308929690144371568755687


class LeadingZeroPrivateAndPublicKeySrpClient(SrpClient):
    def generate_private_key(self):
        return 70997313118674976963008287637113704817


class LeadingZeroPrivateKeySrpServer(SrpServer):
    def generate_private_key(self):
        return 292137137271783308929690144371568755687


class LeadingZeroPrivateAndPublicKeySrpServer(SrpServer):
    def generate_private_key(self):
        return 70997313118674976963008287637113704817


@pytest.mark.parametrize(
    "server_cls, client_cls",
    [
        (SrpServer, SrpClient),
        (ZeroSaltSrpServer, SrpClient),
        (SrpServer, LeadingZeroPrivateKeySrpClient),
        (SrpServer, LeadingZeroPrivateAndPublicKeySrpClient),
        (ZeroSaltSrpServer, LeadingZeroPrivateAndPublicKeySrpClient),
        (
            LeadingZeroPrivateAndPublicKeySrpServer,
            LeadingZeroPrivateAndPublicKeySrpClient,
        ),
        (LeadingZeroPrivateKeySrpServer, LeadingZeroPrivateAndPublicKeySrpClient),
    ],
)
def test_1(server_cls, client_cls):
    # step M1

    # step M2
    setup_code = "123-45-678"  # transmitted on second channel
    server: SrpServer = server_cls("Pair-Setup", setup_code)
    server_pub_key = server.get_public_key_bytes()
    server_salt = server.get_salt()

    # step M3
    client: SrpClient = client_cls("Pair-Setup", setup_code)
    client.set_salt(server_salt)
    client.set_server_public_key(server_pub_key)

    client_pub_key = client.get_public_key_bytes()
    clients_proof_bytes = client.get_proof_bytes()

    # step M4
    server.set_client_public_key(client_pub_key)
    server.get_shared_secret()
    assert server.verify_clients_proof_bytes(clients_proof_bytes) is True
    servers_proof = server.get_proof_bytes(clients_proof_bytes)

    # step M5
    assert client.verify_servers_proof_bytes(servers_proof) is True, (
        f"proof mismatch: server_key:{server.b} client_key:{client.a} server_salt:{server.salt}"
    )



#! env python

import json
import plistlib
import textwrap

from aiohomekit.uuid import normalize_uuid

plist = "/Applications/HomeKit Accessory Simulator.app/Contents/Frameworks/HAPAccessoryKit.framework/Versions/A/Resources/default.metadata.plist"
with open(plist, "rb") as fp:
    data = plistlib.load(fp)


enrichment = {
    "00000120-0000-1000-8000-0026BB765291": {"struct": ".structs.StreamingStatus"},
    "00000117-0000-1000-8000-0026BB765291": {"struct": ".structs.SelectedRTPStreamConfiguration"},
    "00000116-0000-1000-8000-0026BB765291": {
        "struct": ".structs.SupportedRTPConfiguration",
        "array": True,
    },
    "00000115-0000-1000-8000-0026BB765291": {"struct": ".structs.SupportedAudioStreamConfiguration"},
    "00000114-0000-1000-8000-0026BB765291": {"struct": ".structs.SupportedVideoStreamConfiguration"},
}


characteristics = {}

for char in data.get("Characteristics", []):
    name = char["Name"].replace(".", "_").replace(" ", "_").upper()

    c = characteristics[char["UUID"]] = {
        "name": name,
        "description": char["Name"],
    }

    if char.get("Properties"):
        c["perms"] = []
        for perm in char.get("Properties"):
            if perm == "read":
                c["perms"].append("pr")
            if perm == "write":
                c["perms"].append("pw")
            if perm == "cnotify":
                c["perms"].append("ev")

    if char.get("Format"):
        c["format"] = char["Format"]

    if char.get("Unit"):
        c["unit"] = char["Unit"]

    if "Constraints" not in char:
        continue

    constraints = char["Constraints"]

    if constraints.get("MaximumValue"):
        c["max_value"] = constraints["MaximumValue"]
    if constraints.get("MaximumValue"):
        c["min_value"] = constraints["MinimumValue"]
    if constraints.get("StepValue"):
        c["step_value"] = constraints["StepValue"]


with open("aiohomekit/model/characteristics/data.py", "w") as fp:
    fp.write("# AUTOGENERATED, DO NOT EDIT\n\n")

    for char in enrichment.values():
        if "struct" in char:
            imp, frm = char["struct"].rsplit(".", 1)
            fp.write(f"from {imp} import {frm}\n")
    fp.write("\n\n")

    fp.write("characteristics = {\n")

    for char_uuid, char in characteristics.items():
        name = json.dumps(char["name"])
        description = json.dumps(char["description"])
        perms = json.dumps(char["perms"])
        format = json.dumps(char["format"])

        fp.write(f'    "{char_uuid}": {{\n')
        fp.write(f'        "name": {name},\n')
        fp.write(f'        "description": {description},\n')
        fp.write(f'        "perms": {perms},\n')
        fp.write(f'        "format": {format},\n')

        struct = enrichment.get(char_uuid, {}).get("struct")
        if struct:
            _, frm = struct.rsplit(".", 1)
            fp.write(f'        "struct": {frm},\n')

        array = enrichment.get(char_uuid, {}).get("array")
        if array:
            fp.write(f'        "array": {array},\n')

        if "unit" in char:
            unit = json.dumps(char["unit"])
            fp.write(f'        "unit": {unit},\n')

        if "max_value" in char:
            max_value = json.dumps(char["max_value"])
            fp.write(f'        "max_value": {max_value},\n')

        if "min_value" in char:
            min_value = json.dumps(char["min_value"])
            fp.write(f'        "min_value": {min_value},\n')

        if "step_value" in char:
            step_value = json.dumps(char["step_value"])
            fp.write(f'        "min_step": {step_value},\n')

        fp.write("    },\n")
        pass

    fp.write("}\n")


for serv in data.get("Services", []):
    name = serv["Name"].replace(" ", "_").upper()
    short = normalize_uuid(serv["UUID"])
    print(f'{name} = "{short}"')


services = {}

for serv in data.get("Services", []):
    name = serv["Name"].replace(" ", "_").upper()

    s = services[serv["UUID"]] = {
        "name": name,
        "description": serv["Name"],
        "required": serv.get("RequiredCharacteristics", []),
        "optional": serv.get("OptionalCharacteristics", []),
    }


with open("aiohomekit/model/services/data.py", "w") as fp:
    fp.write(
        textwrap.dedent("""
    # AUTOGENERATED, DO NOT EDIT

    services =
    """).strip()
    )
    fp.write(" " + json.dumps(services, indent=4))
    fp.write("\n")



import asyncio
import errno
import logging
import os
import socket
import tempfile
import threading
from unittest import mock
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from zeroconf import DNSCache, SignalRegistrationInterface

from aiohomekit import Controller
from aiohomekit.controller.ip import IpPairing
from aiohomekit.model import Accessory
from aiohomekit.model.characteristics import CharacteristicsTypes
from aiohomekit.model.services import ServicesTypes

from tests.accessoryserver import AccessoryServer


def _get_test_socket() -> socket.socket:
    """Create a socket to test binding ports."""
    test_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    test_socket.setblocking(False)
    test_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    return test_socket


def port_ready(port: int) -> bool:
    try:
        _get_test_socket().bind(("127.0.0.1", port))
    except OSError as e:
        if e.errno == errno.EADDRINUSE:
            return True

    return False


def next_available_port() -> int:
    for port in range(51842, 53842):
        if not port_ready(port):
            return port

    raise RuntimeError("No available ports")


async def wait_for_server_online(port: int):
    for i in range(100):
        if port_ready(port):
            break
        await asyncio.sleep(0.025)


class AsyncServiceBrowserStub:
    types = [
        "_hap._tcp.local.",
        "_hap._udp.local.",
    ]

    def __init__(self):
        self._handlers = []
        self.service_state_changed = SignalRegistrationInterface(self._handlers)


@pytest.fixture
def mock_asynczeroconf():
    """Mock zeroconf."""

    with patch("aiohomekit.zeroconf.AsyncServiceBrowser", AsyncServiceBrowserStub):
        with patch("aiohomekit.zeroconf.AsyncZeroconf") as mock_zc:
            zc = mock_zc.return_value
            zc.async_register_service = AsyncMock()
            zc.async_close = AsyncMock()
            zeroconf = MagicMock(name="zeroconf_mock")
            zeroconf.cache = DNSCache()
            zeroconf.async_wait_for_start = AsyncMock()
            zeroconf.listeners = [AsyncServiceBrowserStub()]
            zc.zeroconf = zeroconf
            yield zc


@pytest.fixture
async def controller_and_unpaired_accessory(request, mock_asynczeroconf, id_factory):
    available_port = next_available_port()

    config_file = tempfile.NamedTemporaryFile(delete=False)
    config_file.write(
        b"""{
        "accessory_ltpk": "7986cf939de8986f428744e36ed72d86189bea46b4dcdc8d9d79a3e4fceb92b9",
        "accessory_ltsk": "3d99f3e959a1f93af4056966f858074b2a1fdec1c5fd84a51ea96f9fa004156a",
        "accessory_pairing_id": "12:34:56:00:01:0A",
        "accessory_pin": "031-45-154",
        "c#": 1,
        "category": "Lightbulb",
        "host_ip": "127.0.0.1",
        "host_port": %port%,
        "name": "unittestLight",
        "unsuccessful_tries": 0
    }""".replace(b"%port%", str(available_port).encode("utf-8"))
    )
    config_file.close()

    httpd = AccessoryServer(config_file.name, None)
    accessory = Accessory.create_with_info(
        id_factory(), "Testlicht", "lusiardi.de", "Demoserver", "0001", "0.1"
    )
    lightBulbService = accessory.add_service(ServicesTypes.LIGHTBULB)
    lightBulbService.add_char(CharacteristicsTypes.ON, value=False)
    httpd.add_accessory(accessory)

    t = threading.Thread(target=httpd.serve_forever)
    t.start()

    await wait_for_server_online(available_port)

    controller = Controller(async_zeroconf_instance=mock_asynczeroconf)

    with mock.patch.object(controller, "load_data", lambda x: None):
        with mock.patch("aiohomekit.__main__.Controller") as c:
            c.return_value = controller
            yield controller, available_port

    os.unlink(config_file.name)

    def _shutdown():
        httpd.shutdown()
        t.join()

    loop = asyncio.get_running_loop()
    asyncio.ensure_future(loop.run_in_executor(None, _shutdown))


@pytest.fixture
async def controller_and_paired_accessory(request, mock_asynczeroconf, id_factory):
    available_port = next_available_port()

    config_file = tempfile.NamedTemporaryFile(delete=False)
    data = b"""{
        "accessory_ltpk": "7986cf939de8986f428744e36ed72d86189bea46b4dcdc8d9d79a3e4fceb92b9",
        "accessory_ltsk": "3d99f3e959a1f93af4056966f858074b2a1fdec1c5fd84a51ea96f9fa004156a",
        "accessory_pairing_id": "12:34:56:00:01:0A",
        "accessory_pin": "031-45-154",
        "c#": 1,
        "category": "Lightbulb",
        "host_ip": "127.0.0.1",
        "host_port": %port%,
        "name": "unittestLight",
        "peers": {
            "decc6fa3-de3e-41c9-adba-ef7409821bfc": {
                "admin": true,
                "key": "d708df2fbf4a8779669f0ccd43f4962d6d49e4274f88b1292f822edc3bcf8ed8"
            }
        },
        "unsuccessful_tries": 0
    }""".replace(b"%port%", str(available_port).encode("utf-8"))

    config_file.write(data)
    config_file.close()

    httpd = AccessoryServer(config_file.name, None)
    accessory = Accessory.create_with_info(
        id_factory(), "Testlicht", "lusiardi.de", "Demoserver", "0001", "0.1"
    )
    lightBulbService = accessory.add_service(ServicesTypes.LIGHTBULB)
    lightBulbService.add_char(CharacteristicsTypes.ON, value=False)
    httpd.add_accessory(accessory)

    t = threading.Thread(target=httpd.serve_forever)
    t.start()

    await wait_for_server_online(available_port)

    controller_file = tempfile.NamedTemporaryFile(delete=False)
    controller_file.write(
        b"""{
        "alias": {
            "Connection": "IP",
            "iOSDeviceLTPK": "d708df2fbf4a8779669f0ccd43f4962d6d49e4274f88b1292f822edc3bcf8ed8",
            "iOSPairingId": "decc6fa3-de3e-41c9-adba-ef7409821bfc",
            "AccessoryLTPK": "7986cf939de8986f428744e36ed72d86189bea46b4dcdc8d9d79a3e4fceb92b9",
            "AccessoryPairingID": "12:34:56:00:01:0A",
            "AccessoryPort": %port%,
            "AccessoryIP": "127.0.0.1",
            "iOSDeviceLTSK": "fa45f082ef87efc6c8c8d043d74084a3ea923a2253e323a7eb9917b4090c2fcc"
        }
    }""".replace(b"%port%", str(available_port).encode("utf-8"))
    )
    controller_file.close()

    controller = Controller(
        async_zeroconf_instance=mock_asynczeroconf,
    )

    async with controller:
        controller.load_data(controller_file.name)
        config_file.close()

        with mock.patch.object(controller, "load_data", lambda x: None):
            with mock.patch("aiohomekit.__main__.Controller") as c:
                c.return_value = controller
                yield controller

    os.unlink(config_file.name)
    os.unlink(controller_file.name)

    def _shutdown():
        httpd.shutdown()
        t.join()

    loop = asyncio.get_running_loop()
    asyncio.ensure_future(loop.run_in_executor(None, _shutdown))


@pytest.fixture
async def pairing(controller_and_paired_accessory):
    pairing = controller_and_paired_accessory.aliases["alias"]
    yield pairing
    try:
        await pairing.close()
    except asyncio.CancelledError:
        pass


@pytest.fixture
async def pairings(request, controller_and_paired_accessory):
    """Returns a pairing of pairngs."""
    left = controller_and_paired_accessory.aliases["alias"]

    right = IpPairing(left.controller, left.pairing_data)

    yield (left, right)

    try:
        await asyncio.shield(right.close())
    except asyncio.CancelledError:
        pass


@pytest.fixture(autouse=True)
def configure_test_logging(caplog):
    caplog.set_level(logging.DEBUG)


@pytest.fixture()
def id_factory():
    id_counter = 0

    def _get_id():
        nonlocal id_counter
        id_counter += 1
        return id_counter

    yield _get_id



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Helpers for detecing homekit devices via zeroconf."""

from __future__ import annotations

import asyncio
import logging
from abc import abstractmethod
from collections.abc import AsyncIterable
from dataclasses import dataclass

from zeroconf import (
    BadTypeInNameException,
    DNSPointer,
    IPVersion,
    ServiceListener,
    ServiceStateChange,
    Zeroconf,
    current_time_millis,
)
from zeroconf.asyncio import AsyncServiceBrowser, AsyncServiceInfo, AsyncZeroconf

from aiohomekit.characteristic_cache import CharacteristicCacheType
from aiohomekit.controller.abstract import (
    AbstractController,
    AbstractDiscovery,
    AbstractPairing,
)
from aiohomekit.exceptions import AccessoryNotFoundError, TransportNotSupportedError
from aiohomekit.model import Categories
from aiohomekit.model.feature_flags import FeatureFlags
from aiohomekit.model.status_flags import StatusFlags

from .utils import async_create_task

HAP_TYPE_TCP = "_hap._tcp.local."
HAP_TYPE_UDP = "_hap._udp.local."
CLASS_IN = 1
TYPE_PTR = 12

_TIMEOUT_MS = 3000

logger = logging.getLogger(__name__)


@dataclass(slots=True)
class HomeKitService:
    name: str
    id: str
    model: str
    feature_flags: FeatureFlags
    status_flags: StatusFlags
    config_num: int
    state_num: int
    category: Categories
    protocol_version: str

    type: str

    address: str
    addresses: list[str]
    port: int

    @classmethod
    def from_service_info(cls, service: AsyncServiceInfo) -> HomeKitService:
        if not (addresses := service.ip_addresses_by_version(IPVersion.All)):
            raise ValueError("Invalid HomeKit Zeroconf record: Missing address")

        address: str | None = None
        #
        # Zeroconf addresses are guaranteed to be returned in LIFO (last in, first out)
        # order with IPv4 addresses first and IPv6 addresses second.
        #
        # This means the first address will always be the most recently added
        # address of the given IP version.
        #
        valid_addresses = [
            str(ip_addr) for ip_addr in addresses if not ip_addr.is_link_local and not ip_addr.is_unspecified
        ]
        if not valid_addresses:
            raise ValueError("Invalid HomeKit Zeroconf record: Missing non-link-local or unspecified address")
        address = valid_addresses[0]
        props = {k.lower(): v for k, v in service.decoded_properties.items() if v is not None}
        if "id" not in props:
            raise ValueError("Invalid HomeKit Zeroconf record: Missing device ID")

        return cls(
            name=service.name.removesuffix(f".{service.type}"),
            id=props["id"].lower(),
            model=props.get("md", ""),
            config_num=int(props.get("c#", 0)),
            state_num=int(props.get("s#", 0)),
            feature_flags=FeatureFlags(int(props.get("ff", 0))),
            status_flags=StatusFlags(int(props.get("sf", 0))),
            category=Categories(int(props.get("ci", 1))),
            protocol_version=props.get("pv", "1.0"),
            type=service.type,
            address=address,
            addresses=valid_addresses,
            port=service.port,
        )


class ZeroconfServiceListener(ServiceListener):
    """An empty service listener."""

    def add_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        """A service has been added."""

    def remove_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        """A service has been removed."""

    def update_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        """A service has been updated."""


def find_brower_for_hap_type(azc: AsyncZeroconf, hap_type: str) -> AsyncServiceBrowser:
    for browser in azc.zeroconf.listeners:
        if not isinstance(browser, AsyncServiceBrowser):
            continue
        if hap_type not in browser.types:
            continue
        return browser

    raise TransportNotSupportedError(f"There is no zeroconf browser for {hap_type}")


class ZeroconfDiscovery(AbstractDiscovery):
    description: HomeKitService

    def __init__(self, description: HomeKitService):
        self.description = description

    def _update_from_discovery(self, description: HomeKitService):
        self.description = description


class ZeroconfPairing(AbstractPairing):
    description: HomeKitService

    def _async_endpoint_changed(self) -> None:
        """The IP and/or port of the accessory has changed."""

    def _async_description_update(self, description: HomeKitService | None) -> None:
        old_description = self.description

        super()._async_description_update(description)

        if not description:
            return

        endpoint_changed = False
        if not old_description:
            logger.debug("%s: Device rediscovered", self.id)
            endpoint_changed = True
        elif old_description.address != description.address:
            logger.debug(
                "%s: Device IP changed from %s to %s",
                self.id,
                old_description.address,
                description.address,
            )
            endpoint_changed = True
        elif old_description.port != description.port:
            logger.debug(
                "%s: Device port changed from %s to %s",
                self.id,
                old_description.port,
                description.port,
            )
            endpoint_changed = True

        if endpoint_changed:
            self._async_endpoint_changed()


class ZeroconfController(AbstractController):
    """
    Base class for HAP protocols that rely on Zeroconf discovery.
    """

    hap_type: str

    def __init__(
        self,
        char_cache: CharacteristicCacheType,
        zeroconf_instance: AsyncZeroconf,
    ):
        super().__init__(char_cache)
        self._async_zeroconf_instance = zeroconf_instance
        self._waiters: dict[str, list[asyncio.Future]] = {}
        self._resolve_later: dict[str, asyncio.TimerHandle] = {}
        self._loop = asyncio.get_running_loop()
        self._running = True

    async def async_start(self):
        zc = self._async_zeroconf_instance.zeroconf
        if not zc:
            return self

        self._browser = find_brower_for_hap_type(self._async_zeroconf_instance, self.hap_type)
        self._browser.service_state_changed.register_handler(self._handle_service)
        await self._async_update_from_cache(zc)

        return self

    async def _async_update_from_cache(self, zc: Zeroconf) -> None:
        """Load the records from the cache."""
        tasks: list[asyncio.Task] = []
        now = current_time_millis()
        for record in self._async_get_ptr_records(zc):
            try:
                info = AsyncServiceInfo(self.hap_type, record.alias)
            except BadTypeInNameException as ex:
                logger.debug("Ignoring record with bad type in name: %s: %s", record.alias, ex)
                continue
            if info.load_from_cache(zc, now):
                self._async_handle_loaded_service_info(info)
            else:
                tasks.append(self._async_handle_service(info))

        if tasks:
            await asyncio.gather(*tasks)

    def _async_get_ptr_records(self, zc: Zeroconf) -> list[DNSPointer]:
        """Return all PTR records for the HAP type."""
        return zc.cache.async_all_by_details(self.hap_type, TYPE_PTR, CLASS_IN)

    def _handle_service(
        self,
        zeroconf: Zeroconf,
        service_type: str,
        name: str,
        state_change: ServiceStateChange,
    ) -> None:
        if service_type != self.hap_type:
            return

        if state_change == ServiceStateChange.Removed:
            if cancel := self._resolve_later.pop(name, None):
                cancel.cancel()
            return

        if name in self._resolve_later:
            # We already have a timer to resolve this service, so ignore this
            # callback.
            return

        try:
            info = AsyncServiceInfo(service_type, name)
        except BadTypeInNameException as ex:
            logger.debug("Ignoring record with bad type in name: %s: %s", name, ex)
            return

        self._resolve_later[name] = self._loop.call_at(
            self._loop.time() + 0.5, self._async_resolve_later, name, info
        )

    def _async_resolve_later(self, name: str, info: AsyncServiceInfo) -> None:
        """Resolve a host later."""
        # As soon as we get a callback, we can remove the _resolve_later
        # so the next time we get a callback, we can resolve the service
        # again if needed which ensures the TTL is respected.
        self._resolve_later.pop(name, None)

        if not self._running:
            return

        if info.load_from_cache(self._async_zeroconf_instance.zeroconf):
            self._async_handle_loaded_service_info(info)
        else:
            async_create_task(self._async_handle_service(info))

    async def async_stop(self):
        """Stop the controller."""
        self._running = False
        self._browser.service_state_changed.unregister_handler(self._handle_service)
        while self._resolve_later:
            _, cancel = self._resolve_later.popitem()
            cancel.cancel()

    async def async_find(self, device_id: str, timeout: float = 10.0) -> ZeroconfDiscovery:
        device_id = device_id.lower()

        if discovery := self.discoveries.get(device_id):
            return discovery

        waiters = self._waiters.setdefault(device_id, [])
        waiter = self._loop.create_future()
        waiters.append(waiter)
        cancel_timeout = self._loop.call_later(timeout, self._async_on_timeout, waiter)

        try:
            if discovery := await waiter:
                return discovery
        except asyncio.TimeoutError:
            raise AccessoryNotFoundError(f"Accessory with device id {device_id} not found")
        finally:
            cancel_timeout.cancel()

    async def async_reachable(self, device_id: str, timeout: float = 10.0) -> bool:
        """Check if a device is reachable.

        This method will return True if the device is reachable, False if it is not.

        Typically A/AAAA records have a TTL of 120 seconds which means we may
        see the device as reachable for up to 120 seconds after it has been
        removed from the network if it does not send a goodbye packet.
        """
        try:
            discovery = await self.async_find(device_id, timeout)
        except AccessoryNotFoundError:
            return False
        alias = f"{discovery.description.name}.{self.hap_type}"
        info = AsyncServiceInfo(self.hap_type, alias)
        zc = self._async_zeroconf_instance.zeroconf
        return info.load_from_cache(zc) or await info.async_request(zc, _TIMEOUT_MS)

    def _async_on_timeout(self, future: asyncio.Future) -> None:
        if not future.done():
            future.set_exception(asyncio.TimeoutError())

    async def async_discover(self) -> AsyncIterable[ZeroconfDiscovery]:
        for device in self.discoveries.values():
            yield device

    async def _async_handle_service(self, info: AsyncServiceInfo):
        """Add a device that became visible via zeroconf."""
        # AsyncServiceInfo already tries 3x
        await info.async_request(self._async_zeroconf_instance.zeroconf, _TIMEOUT_MS)
        self._async_handle_loaded_service_info(info)

    def _async_handle_loaded_service_info(self, info: AsyncServiceInfo) -> None:
        """Handle a service info that was discovered via zeroconf."""
        try:
            description = HomeKitService.from_service_info(info)
        except ValueError as e:
            logger.debug("%s: Not a valid homekit device: %s", info.name, e)
            return

        if discovery := self.discoveries.get(description.id):
            discovery._update_from_discovery(description)
        else:
            discovery = self.discoveries[description.id] = self._make_discovery(description)

        discovery = self.discoveries[description.id] = self._make_discovery(description)

        if pairing := self.pairings.get(description.id):
            logger.debug(
                "%s: Notifying pairing of description update: %s",
                description.id,
                description,
            )
            pairing._async_description_update(description)

        if waiters := self._waiters.pop(description.id, None):
            for waiter in waiters:
                if not waiter.cancelled() and not waiter.done():
                    waiter.set_result(discovery)

    @abstractmethod
    def _make_discovery(self, description: HomeKitService) -> AbstractDiscovery:
        pass



from aiohomekit import Controller
from aiohomekit.controller.ip import IpDiscovery, IpPairing
from aiohomekit.model.categories import Categories
from aiohomekit.zeroconf import HomeKitService


async def test_pair(controller_and_unpaired_accessory: tuple[Controller, int]):
    controller, port = controller_and_unpaired_accessory

    discovery = IpDiscovery(
        controller,
        HomeKitService(
            name="Test",
            id="00:01:02:03:04:05",
            model="Test",
            feature_flags=0,
            status_flags=1,
            config_num=0,
            state_num=0,
            category=Categories.OTHER,
            protocol_version="1.0",
            type="_hap._tcp.local",
            address="127.0.0.1",
            addresses=["127.0.0.1"],
            port=port,
        ),
    )

    finish_pairing = await discovery.async_start_pairing("alias")
    pairing = await finish_pairing("031-45-154")

    assert isinstance(pairing, IpPairing)

    assert await pairing.get_characteristics([(1, 9)]) == {
        (1, 9): {"value": False},
    }


async def test_identify(controller_and_unpaired_accessory: tuple[Controller, int]):
    controller, port = controller_and_unpaired_accessory

    discovery = IpDiscovery(
        controller,
        HomeKitService(
            name="Test",
            id="00:01:02:03:04:05",
            model="Test",
            feature_flags=0,
            status_flags=0,
            config_num=0,
            state_num=0,
            category=Categories.OTHER,
            protocol_version="1.0",
            type="_hap._tcp.local",
            address="127.0.0.1",
            addresses=["127.0.0.1"],
            port=port,
        ),
    )

    identified = await discovery.async_identify()
    assert identified is True



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from aiohomekit.model import Accessory
from aiohomekit.model.characteristics import (
    CharacteristicsTypes,
    InputEventValues,
    RemoteKeyValues,
)
from aiohomekit.model.feature_flags import FeatureFlags
from aiohomekit.model.services import ServicesTypes
from aiohomekit.utils import (
    clamp_enum_to_char,
    domain_supported,
    domain_to_name,
    pair_with_auth,
)
from aiohomekit.uuid import normalize_uuid, shorten_uuid


def test_normalize_short_uuid():
    assert normalize_uuid("121") == "00000121-0000-1000-8000-0026BB765291"


def test_normalize_uuid():
    assert normalize_uuid("00000121-0000-1000-8000-0026BB765291") == "00000121-0000-1000-8000-0026BB765291"


def test_normalize_uuid_coap_short():
    assert normalize_uuid("45E5011ECB4000A80FF2603DE") == "00000004-5E50-11EC-B400-0A80FF2603DE"


def test_normalize_invalid_uuid():
    with pytest.raises(Exception):
        normalize_uuid("NOT_A_VALID_UUID")


def test_shorten_uuid():
    assert shorten_uuid("00000121-0000-1000-8000-0026BB765291") == "121"


def test_shorten_vendor_uuid():
    assert shorten_uuid("00000121-0000-1000-8000-AAAAAAAAAAAA") == "00000121-0000-1000-8000-AAAAAAAAAAAA"


def test_shorten_invalid_uuid():
    with pytest.raises(Exception):
        shorten_uuid("NOT_A_VALID_UUID")


def test_clamp_enum_valid_vals(id_factory):
    a = Accessory(id_factory())
    tv_service = a.add_service(service_type=ServicesTypes.TELEVISION)
    char = tv_service.add_char(
        CharacteristicsTypes.REMOTE_KEY,
        valid_values=[RemoteKeyValues.PLAY_PAUSE],
        min_value=None,
        max_value=None,
    )

    valid_vals = clamp_enum_to_char(RemoteKeyValues, char)
    assert valid_vals == {RemoteKeyValues.PLAY_PAUSE}


def test_clamp_enum_min_max(id_factory):
    a = Accessory(id_factory())
    tv_service = a.add_service(service_type=ServicesTypes.TELEVISION)
    char = tv_service.add_char(
        CharacteristicsTypes.REMOTE_KEY,
        valid_values=None,
        min_value=RemoteKeyValues.PLAY_PAUSE,
        max_value=RemoteKeyValues.PLAY_PAUSE,
    )

    valid_vals = clamp_enum_to_char(RemoteKeyValues, char)

    assert valid_vals == {RemoteKeyValues.PLAY_PAUSE}


def test_clamp_enum_min_max_single_press(id_factory):
    a = Accessory(id_factory())
    tv_service = a.add_service(service_type=ServicesTypes.STATELESS_PROGRAMMABLE_SWITCH)
    char = tv_service.add_char(
        CharacteristicsTypes.INPUT_EVENT,
        valid_values=None,
        min_value=InputEventValues.SINGLE_PRESS,
        max_value=InputEventValues.SINGLE_PRESS,
    )

    valid_vals = clamp_enum_to_char(InputEventValues, char)

    assert valid_vals == {InputEventValues.SINGLE_PRESS}


def test_clamp_enum_min_max_unclamped_button_press(id_factory):
    a = Accessory(id_factory())
    tv_service = a.add_service(service_type=ServicesTypes.STATELESS_PROGRAMMABLE_SWITCH)
    char = tv_service.add_char(
        CharacteristicsTypes.INPUT_EVENT,
    )

    valid_vals = clamp_enum_to_char(InputEventValues, char)

    assert valid_vals == {
        InputEventValues.SINGLE_PRESS,
        InputEventValues.DOUBLE_PRESS,
        InputEventValues.LONG_PRESS,
    }


def test_pair_with_auth():
    # Only for FeatureFlags.SUPPORTS_APPLE_AUTHENTICATION_COPROCESSOR
    assert pair_with_auth(FeatureFlags(1)) is True
    assert pair_with_auth(FeatureFlags(2)) is False
    assert pair_with_auth(FeatureFlags(0)) is False
    assert pair_with_auth(FeatureFlags(4)) is False


def test_domain_to_name():
    assert domain_to_name("Bar._hap._tcp.local.") == "Bar"
    assert domain_to_name("Foo's Library._music._tcp.local.") == "Foo's Library"


def test_domain_supported():
    assert domain_supported("Bar._hap._tcp.local.")
    assert not domain_supported("Bar._music._tcp.local.")



import asyncio
from datetime import timedelta
from typing import Any
from unittest import mock

import pytest

from aiohomekit.controller.ip.pairing import IpPairing
from aiohomekit.exceptions import AccessoryDisconnectedError
from aiohomekit.model import Transport
from aiohomekit.protocol.statuscodes import HapStatusCode


async def test_list_accessories(pairing: IpPairing):
    accessories = await pairing.list_accessories_and_characteristics()
    assert accessories[0]["aid"] == 1
    assert accessories[0]["services"][0]["iid"] == 1

    char = accessories[0]["services"][0]["characteristics"][0]

    assert char["description"] == "Identify"
    assert char["iid"] == 2
    assert char["format"] == "bool"
    assert char["perms"] == ["pw"]
    assert char["type"] == "00000014-0000-1000-8000-0026BB765291"


async def test_get_characteristics(pairing: IpPairing):
    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics[(1, 9)] == {"value": False}


async def test_duplicate_get_characteristics(pairing):
    characteristics = await pairing.get_characteristics([(1, 9), (1, 9)])
    assert characteristics[(1, 9)] == {"value": False}


async def test_get_characteristics_after_failure(pairing: IpPairing):
    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics[(1, 9)] == {"value": False}

    pairing.connection.transport.close()
    await asyncio.sleep(0)
    assert not pairing.connection.is_connected
    assert not pairing.is_available

    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics[(1, 9)] == {"value": False}


async def test_reconnect_soon_after_disconnected(pairing: IpPairing):
    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics[(1, 9)] == {"value": False}

    assert pairing.connection.is_connected
    assert pairing.is_available

    pairing.connection.transport.close()
    await asyncio.sleep(0)
    assert not pairing.connection.is_connected
    assert not pairing.is_available

    # Ensure we can safely call multiple times
    pairing._async_description_update(None)
    pairing._async_description_update(None)
    pairing._async_description_update(None)

    await asyncio.sleep(0)  # ensure the callback has a chance to run and create _connector
    await asyncio.wait_for(pairing.connection._connector, timeout=0.5)
    assert pairing.connection.is_connected

    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics[(1, 9)] == {"value": False}


async def test_reconnect_soon_after_device_is_offline_for_a_bit(pairing: IpPairing):
    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics[(1, 9)] == {"value": False}

    assert pairing.connection.is_connected
    assert pairing.is_available

    with mock.patch(
        "aiohomekit.controller.ip.connection.HomeKitConnection._connect_once",
        side_effect=asyncio.TimeoutError,
    ):
        pairing.connection.transport.close()
        await asyncio.sleep(0)
        assert not pairing.connection.is_connected
        assert not pairing.is_available

        for _ in range(3):
            pairing._async_description_update(None)
            # ensure the callback has a chance to run and create _connector
            await asyncio.sleep(0)
            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(asyncio.shield(pairing.connection._connector), timeout=0.2)
            assert not pairing.connection.is_connected

    pairing._async_description_update(None)
    await asyncio.wait_for(pairing.connection._connector, timeout=0.5)
    assert pairing.connection.is_connected
    assert pairing.is_available

    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics[(1, 9)] == {"value": False}


async def test_reconnect_soon_on_device_reboot(pairing: IpPairing):
    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics[(1, 9)] == {"value": False}

    assert pairing.connection.is_connected
    assert pairing.is_available

    with mock.patch(
        "aiohomekit.controller.ip.connection.HomeKitConnection._connect_once",
        side_effect=asyncio.TimeoutError,
    ):
        pairing.connection.protocol.connection_lost(OSError("Connection reset by peer"))

    await asyncio.sleep(0)
    assert not pairing.connection.is_connected
    assert not pairing.is_available
    await asyncio.wait_for(pairing.connection._connector, timeout=0.5)
    assert pairing.connection.is_connected
    assert pairing.is_available

    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics[(1, 9)] == {"value": False}


async def test_put_characteristics(pairing: IpPairing):
    characteristics = await pairing.put_characteristics([(1, 9, True)])

    assert characteristics == {}

    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics[(1, 9)] == {"value": True}


async def test_put_characteristics_cancelled(pairing: IpPairing):
    characteristics = await pairing.put_characteristics([(1, 9, True)])
    characteristics = await pairing.get_characteristics([(1, 9)])

    with mock.patch.object(pairing.connection.transport, "writelines"):
        task = asyncio.create_task(pairing.put_characteristics([(1, 9, False)]))
        await asyncio.sleep(0)
        for future in pairing.connection.protocol.result_cbs:
            future.cancel()
        await asyncio.sleep(0)
        with pytest.raises(asyncio.CancelledError):
            await task

    # We should wait a few seconds to see if the
    # connection can be re-established and the write can be
    # completed. But this is not currently possible because
    # we do not wait for the connection to be re-established
    # before we try to write the data. When we implement
    # reconnection we should remove this pytest.raises
    # and the sleep below.
    with pytest.raises(AccessoryDisconnectedError):
        await pairing.get_characteristics([(1, 9)])

    await asyncio.sleep(0)
    characteristics = await pairing.get_characteristics([(1, 9)])
    assert characteristics[(1, 9)] == {"value": True}


async def test_put_characteristics_callbacks(pairing: IpPairing):
    events = []

    def process_new_events(new_values_dict: dict[tuple[int, int], dict[str, Any]]) -> None:
        events.append(new_values_dict)

    pairing.dispatcher_connect(process_new_events)
    assert events == []

    characteristics = await pairing.put_characteristics([(1, 9, True)])
    assert events == [{}, {(1, 9): {"value": True}}]
    assert characteristics == {}

    # Identify is a write only characteristic, so we should not get a callback
    characteristics = await pairing.put_characteristics([(1, 2, True)])
    assert events == [{}, {(1, 9): {"value": True}}]

    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics[(1, 9)] == {"value": True}

    characteristics = await pairing.get_characteristics({(1, 9)})

    assert characteristics[(1, 9)] == {"value": True}


async def test_subscribe(pairing: IpPairing):
    assert pairing.subscriptions == set()

    await pairing.subscribe([(1, 9)])

    assert pairing.subscriptions == {(1, 9)}

    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics == {(1, 9): {"value": False}}


async def test_unsubscribe(pairing: IpPairing):
    await pairing.subscribe([(1, 9)])

    assert pairing.subscriptions == {(1, 9)}

    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics == {(1, 9): {"value": False}}

    await pairing.unsubscribe([(1, 9)])

    assert pairing.subscriptions == set()

    characteristics = await pairing.get_characteristics([(1, 9)])

    assert characteristics == {(1, 9): {"value": False}}


async def test_dispatcher_connect(pairing: IpPairing):
    assert pairing.listeners == set()

    def callback(x):
        pass

    cancel = pairing.dispatcher_connect(callback)
    assert pairing.listeners == {callback}

    cancel()
    assert pairing.listeners == set()


async def test_receiving_events(pairings):
    """
    Test that can receive events when change happens in another session.

    We set up 2 controllers both with active secure sessions. One
    subscribes and then other does put() calls.

    This test is currently skipped because accessory server doesnt
    support events.
    """
    left: IpPairing = pairings[0]
    right: IpPairing = pairings[1]

    event_value = None
    ev = asyncio.Event()

    def handler(data):
        print(data)
        nonlocal event_value
        event_value = data
        ev.set()

    # Set where to send events
    right.dispatcher_connect(handler)

    # Set what events to get
    await right.subscribe([(1, 9)])

    # Trigger an event by writing a change on the other connection
    await left.put_characteristics([(1, 9, True)])

    # Wait for event to be received for up to 5s
    await asyncio.wait_for(ev.wait(), 5)

    assert event_value == {(1, 9): {"value": True}}


async def test_subscribe_invalid_iid(pairing: IpPairing):
    """
    Test that can get an error when subscribing to an invalid iid.
    """
    result = await pairing.subscribe([(1, 999999)])
    assert result == {
        (1, 999999): {
            "description": "Resource does not exist.",
            "status": HapStatusCode.RESOURCE_NOT_EXIST.value,
        }
    }


async def test_list_pairings(pairing: IpPairing):
    pairings = await pairing.list_pairings()
    assert pairings == [
        {
            "controllerType": "admin",
            "pairingId": "decc6fa3-de3e-41c9-adba-ef7409821bfc",
            "permissions": 1,
            "publicKey": "d708df2fbf4a8779669f0ccd43f4962d6d49e4274f88b1292f822edc3bcf8ed8",
        }
    ]


async def test_add_pairings(pairing: IpPairing):
    await pairing.add_pairing(
        "decc6fa3-de3e-41c9-adba-ef7409821bfe",
        "d708df2fbf4a8779669f0ccd43f4962d6d49e4274f88b1292f822edc3bcf8ed7",
        "User",
    )

    pairings = await pairing.list_pairings()
    assert pairings == [
        {
            "controllerType": "admin",
            "pairingId": "decc6fa3-de3e-41c9-adba-ef7409821bfc",
            "permissions": 1,
            "publicKey": "d708df2fbf4a8779669f0ccd43f4962d6d49e4274f88b1292f822edc3bcf8ed8",
        },
        {
            "controllerType": "regular",
            "pairingId": "decc6fa3-de3e-41c9-adba-ef7409821bfe",
            "permissions": 0,
            "publicKey": "d708df2fbf4a8779669f0ccd43f4962d6d49e4274f88b1292f822edc3bcf8ed7",
        },
    ]


async def test_add_and_remove_pairings(pairing: IpPairing):
    await pairing.add_pairing(
        "decc6fa3-de3e-41c9-adba-ef7409821bfe",
        "d708df2fbf4a8779669f0ccd43f4962d6d49e4274f88b1292f822edc3bcf8ed7",
        "User",
    )

    pairings = await pairing.list_pairings()
    assert pairings == [
        {
            "controllerType": "admin",
            "pairingId": "decc6fa3-de3e-41c9-adba-ef7409821bfc",
            "permissions": 1,
            "publicKey": "d708df2fbf4a8779669f0ccd43f4962d6d49e4274f88b1292f822edc3bcf8ed8",
        },
        {
            "controllerType": "regular",
            "pairingId": "decc6fa3-de3e-41c9-adba-ef7409821bfe",
            "permissions": 0,
            "publicKey": "d708df2fbf4a8779669f0ccd43f4962d6d49e4274f88b1292f822edc3bcf8ed7",
        },
    ]

    await pairing.remove_pairing("decc6fa3-de3e-41c9-adba-ef7409821bfe")

    pairings = await pairing.list_pairings()
    assert pairings == [
        {
            "controllerType": "admin",
            "pairingId": "decc6fa3-de3e-41c9-adba-ef7409821bfc",
            "permissions": 1,
            "publicKey": "d708df2fbf4a8779669f0ccd43f4962d6d49e4274f88b1292f822edc3bcf8ed8",
        }
    ]


async def test_identify(pairing):
    identified = await pairing.identify()
    assert identified is True


async def test_transport_property(pairing: IpPairing):
    assert pairing.transport == Transport.IP


async def test_polling_property(pairing: IpPairing):
    assert pairing.poll_interval == timedelta(seconds=60)


async def test_put_characteristics_invalid_value(pairing: IpPairing):
    aid, iid = (1, 2)
    characteristics = [(aid, iid, 100)]
    status_code = await pairing.put_characteristics(characteristics)
    assert status_code is not None
    assert status_code[(aid, iid)] is not None
    assert status_code[(aid, iid)]["status"] == HapStatusCode.INVALID_VALUE.value


async def test_put_characteristics_boolean_response(pairing: IpPairing):
    """Test handling of malformed response with boolean values (issue #465)."""
    # Mock the connection's put_json to return the malformed response
    # Use the existing accessory ID from the test fixture
    malformed_response = {
        "characteristics": [
            True,  # Boolean instead of dict
            {"iid": 9, "aid": 1, "status": -70402},
            True,  # Another boolean
        ]
    }

    with mock.patch.object(pairing.connection, "put_json", return_value=malformed_response):
        # This should not raise TypeError anymore
        result = await pairing.put_characteristics([(1, 9, True)])

        # Should only process the valid error response
        assert (1, 9) in result
        assert result[(1, 9)]["status"] == -70402
        assert "Unable to communicate" in result[(1, 9)]["description"]



from dataclasses import dataclass, field
from enum import IntEnum

from aiohomekit.tlv8 import TLVStruct, tlv_entry, u8, u64, u128


def test_example_1():
    # Based on 14.1.2 example 1 in R2 spec

    @dataclass
    class DummyStruct(TLVStruct):
        state: u8 = tlv_entry(7)
        message: str = tlv_entry(1)

    raw = b"\x07\x01\x03\x01\x05\x68\x65\x6c\x6c\x6f"

    result = DummyStruct.decode(raw)

    assert result.state == 3
    assert result.message == "hello"

    assert result.encode() == raw


def test_example_2():
    # Based on 14.1.2 example 1 in R2 spec

    @dataclass
    class DummyStruct(TLVStruct):
        state: u8 = tlv_entry(6)
        certificate: str = tlv_entry(9)
        identifier: str = tlv_entry(1)

    raw = b"\x06\x01\x03\x09\xff" + b"a" * 255 + b"\x09\x2d" + b"a" * 45 + b"\x01\x05hello"

    result = DummyStruct.decode(raw)

    assert result.state == 3
    assert result.certificate == "a" * 300
    assert result.identifier == "hello"

    assert result.encode() == raw


def test_ignore_field():
    @dataclass
    class DummyStruct(TLVStruct):
        # fields stored locally for each house boat
        number_of_residents: u8 = field(init=False, default=1)

        # fields pulled from satellite network
        sharks_nearby: u8 = tlv_entry(1)
        days_ago_sharks_ate: u8 = tlv_entry(3)
        sharks_have_coherent_monochromatic_light_sources: u8 = tlv_entry(5)

        def risk_of_shark_attack(self):
            return 100  # the old algorithm was incorrect

    raw = b"\x01\x01\xf0" + b"\x03\x01\x00" + b"\x05\x01\x01"

    result = DummyStruct.decode(raw)

    assert result.number_of_residents == 1
    assert result.sharks_nearby == 240
    assert result.days_ago_sharks_ate == 0
    assert result.sharks_have_coherent_monochromatic_light_sources == 1

    result.number_of_residents = 0

    assert result.encode() == raw


def test_int_enum():
    class FooValues(IntEnum):
        ACTIVE = 1
        INACTIVE = 0

    @dataclass
    class DummyStruct(TLVStruct):
        foo: FooValues = tlv_entry(1)

    result = DummyStruct.decode(b"\x01\x01\x01")
    assert result.foo == FooValues.ACTIVE

    assert DummyStruct(foo=FooValues.ACTIVE).encode() == b"\x01\x01\x01"
    assert DummyStruct(foo=FooValues.INACTIVE).encode() == b"\x01\x01\x00"


def test_int_64b():
    @dataclass
    class DummyStruct(TLVStruct):
        prefix: u64 = tlv_entry(48)

    raw = b"\x30\x08" + b"\xbb\xe8\xec<wS\xc8\xfd"

    result = DummyStruct.decode(raw)

    assert f"{result.prefix:X}" == "FDC853773CECE8BB"

    assert result.encode() == raw


def test_int_uuid():
    @dataclass
    class DummyStruct(TLVStruct):
        dummy_type: u128 = tlv_entry(16)

    raw = b"\x10\x10" + b"\x83\xc3*\xa6yS\xb1N\x90\xdetZ\xb4l:&"

    result = DummyStruct.decode(raw)

    assert f"{result.dummy_type:X}" == "263A6CB45A74DE904EB15379A62AC383"

    assert result.encode() == raw


def test_bytes():
    @dataclass
    class DummyStruct(TLVStruct):
        value: bytes = tlv_entry(160)

    raw = (
        b"\xa0\x12" + b"\x88\x01\x00\x34" + b"\x00\x01\x00\x05\x74\x68\x2f\x74\x63" + b"\x00\x03\x00\x27\x00"
    )

    result = DummyStruct.decode(raw)

    assert result.value == b"\x88\x01\x00\x34\x00\x01\x00\x05\x74\x68\x2f\x74\x63\x00\x03\x00\x27\x00"

    assert result.encode() == raw



from aiohomekit.meshcop import Meshcop


def test_parse_tlv() -> None:
    """Test the TLV parser."""
    dataset_tlv = (
        "0E080000000000010000000300000F35060004001FFFE0020811111111222222220708FDAD70BF"
        "E5AA15DD051000112233445566778899AABBCCDDEEFF030E4F70656E54687265616444656D6F01"
        "0212340410445F2B5CA6F2A93A55CE570A70EFEECB0C0402A0F7F8"
    )

    struct = Meshcop.decode(bytes.fromhex(dataset_tlv))

    assert struct.networkname == "OpenThreadDemo"
    assert struct.channel == 15
    assert struct.panid == 4660
    assert struct.extpanid == bytes.fromhex("1111111122222222")
    assert struct.pskc == bytes.fromhex("445f2b5ca6f2a93a55ce570a70efeecb")
    assert struct.networkkey == bytes.fromhex("00112233445566778899aabbccddeeff")
    assert struct.meshlocalprefix == bytes.fromhex("fdad70bfe5aa15dd")
    assert struct.securitypolicy == bytes.fromhex("02a0f7f8")
    assert struct.activetimestamp == bytes.fromhex("0000000000010000")
    assert struct.channelmask == bytes.fromhex("0004001fffe0")



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

import aiohomekit.hkjson as hkjson


def test_loads_trailing_comma():
    """Test we can decode with a trailing comma."""
    result = hkjson.loads(
        '{"characteristics":[{"aid":10,"iid":12,"value":27.0},{"aid":10,"iid":13,"value":20.5},]}'
    )
    assert result == {
        "characteristics": [
            {"aid": 10, "iid": 12, "value": 27.0},
            {"aid": 10, "iid": 13, "value": 20.5},
        ]
    }


def test_loads_empty_document():
    """Test that empty document raises ValueError instead of lark error."""
    with pytest.raises(ValueError, match="Failed to parse JSON"):
        hkjson.loads("")



from aiohomekit.controller.coap.pdu import (
    OpCode,
    PDUStatus,
    decode_all_pdus,
    decode_pdu,
    encode_all_pdus,
    encode_pdu,
)


def test_encode_without_data():
    req_pdu = encode_pdu(OpCode.CHAR_READ, 0x10, 0x2022, b"")

    assert req_pdu == b"\x00\x03\x10\x22\x20\x00\x00"


def test_encode_with_data():
    req_pdu = encode_pdu(OpCode.CHAR_WRITE, 0x20, 0x1234, b"\x01\x02\x03\x04")

    assert req_pdu == b"\x00\x02\x20\x34\x12\x04\x00\x01\x02\x03\x04"


def test_encode_all_without_data():
    req_pdu = encode_all_pdus(OpCode.CHAR_READ, [0x10, 0x11], [b"", b""])

    assert req_pdu == b"\x00\x03\x00\x10\x00\x00\x00" + b"\x00\x03\x01\x11\x00\x00\x00"


def test_encode_all_with_data():
    req_pdu = encode_all_pdus(OpCode.CHAR_WRITE, [0x12, 0x13], [b"\x88\x88", b"\x99\x99\x99\x99"])

    assert (
        req_pdu == b"\x00\x02\x00\x12\x00\x02\x00\x88\x88" + b"\x00\x02\x01\x13\x00\x04\x00\x99\x99\x99\x99"
    )


def test_decode_without_data():
    res_pdu = b"\x02\x40\x00\x00\x00"
    res_len, res_val = decode_pdu(0x40, res_pdu)

    assert res_len == 0
    assert res_val == b""


def test_decode_with_data():
    res_pdu = b"\x02\x50\x00\x06\x00\x01\x01\x01\x02\x01\x00"
    res_len, res_val = decode_pdu(0x50, res_pdu)

    assert res_len == 6
    assert res_val == b"\x01\x01\x01\x02\x01\x00"


def test_decode_all_without_data():
    res_pdu = b"\x02\x20\x00\x00\x00" + b"\x02\x21\x00\x00\x00"
    res = decode_all_pdus(0x20, res_pdu)

    assert len(res) == 2
    assert isinstance(res[0], bytes)
    assert isinstance(res[1], bytes)
    assert len(res[0]) == 0
    assert len(res[1]) == 0


def test_decode_all_with_data():
    res_pdu = b"\x02\x30\x00\x02\x00\x01\x00" + b"\x02\x31\x00\x02\x00\x02\x00"
    res = decode_all_pdus(0x30, res_pdu)

    assert len(res) == 2
    assert isinstance(res[0], bytes)
    assert isinstance(res[1], bytes)
    assert res[0] == b"\x01\x00"
    assert res[1] == b"\x02\x00"


def test_decode_all_with_single_bad_tid():
    res_pdu = b"\x02\x40\x00\x02\x00\x03\x00" + b"\x02\x99\x00\x02\x00\x04\x00" + b"\x02\x42\x00\x00\x00"
    res = decode_all_pdus(0x40, res_pdu)

    assert len(res) == 3
    assert isinstance(res[0], bytes)
    assert isinstance(res[1], PDUStatus)
    assert isinstance(res[2], bytes)
    assert res[0] == b"\x03\x00"
    assert res[1] == PDUStatus.TID_MISMATCH
    assert len(res[2]) == 0


def test_decode_all_with_single_status_error():
    res_pdu = b"\x02\x50\x00\x00\x00" + b"\x02\x51\x06\x00\x00" + b"\x02\x52\x00\x02\x00\x05\x00"
    res = decode_all_pdus(0x50, res_pdu)

    assert len(res) == 3
    assert isinstance(res[0], bytes)
    assert isinstance(res[1], PDUStatus)
    assert isinstance(res[2], bytes)
    assert len(res[0]) == 0
    assert res[1] == PDUStatus.INVALID_REQUEST
    assert res[2] == b"\x05\x00"


def test_decode_all_with_single_bad_control():
    res_pdu = b"\x02\x50\x00\x00\x00" + b"\x08\x51\x00\x00\x00" + b"\x02\x52\x00\x02\x00\x06\x00"
    res = decode_all_pdus(0x50, res_pdu)

    assert len(res) == 3
    assert isinstance(res[0], bytes)
    assert isinstance(res[1], PDUStatus)
    assert isinstance(res[2], bytes)
    assert len(res[0]) == 0
    assert res[1] == PDUStatus.BAD_CONTROL
    assert res[2] == b"\x06\x00"


def test_decode_with_bad_tid():
    res_pdu = b"\x02\x99\x00\x00\x00"
    res_len, res_val = decode_pdu(0x60, res_pdu)

    assert res_len == 0
    assert res_val == PDUStatus.TID_MISMATCH


def test_decode_with_status_error():
    res_pdu = b"\x02\x99\x06\x00\x00"
    res_len, res_val = decode_pdu(0x99, res_pdu)

    assert res_len == 0
    assert isinstance(res_val, PDUStatus)
    assert res_val == PDUStatus.INVALID_REQUEST


def test_decode_with_bad_control():
    res_pdu = b"\xcc\x99\x00\x00\x00"
    res_len, res_val = decode_pdu(0x99, res_pdu)

    assert res_len == 0
    assert res_val == PDUStatus.BAD_CONTROL



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from aiohomekit.protocol.tlv import TLV, TlvParseException


class TestTLV(unittest.TestCase):
    def test_long_values_1(self):
        val = [
            [TLV.kTLVType_State, TLV.M3],
            [TLV.kTLVType_Certificate, (300 * "a").encode()],
            [TLV.kTLVType_Identifier, b"hello"],
        ]
        res = TLV.decode_bytearray(TLV.encode_list(val))
        self.assertEqual(val, res)

    def test_long_values_2(self):
        val = [
            [TLV.kTLVType_State, TLV.M3],
            [TLV.kTLVType_Certificate, (150 * "a" + 150 * "b").encode()],
            [TLV.kTLVType_Identifier, b"hello"],
        ]
        res = TLV.decode_bytearray(TLV.encode_list(val))
        self.assertEqual(val, res)

    def test_long_values_decode_bytearray_to_list(self):
        example = bytearray.fromhex("060103" + ("09FF" + 255 * "61" + "092D" + 45 * "61") + "010568656c6c6f")
        expected = [
            [6, bytearray(b"\x03")],
            [9, bytearray(300 * b"a")],
            [1, bytearray(b"hello")],
        ]

        data = TLV.decode_bytearray(example)
        self.assertListEqual(data, expected)

    def test_long_values_decode_bytes_to_list(self):
        example = bytes(
            bytearray.fromhex("060103" + ("09FF" + 255 * "61" + "092D" + 45 * "61") + "010568656c6c6f")
        )
        expected = [
            [6, bytearray(b"\x03")],
            [9, bytearray(300 * b"a")],
            [1, bytearray(b"hello")],
        ]

        data = TLV.decode_bytes(example)
        self.assertListEqual(data, expected)

    # def test_long_values_decode_bytearray(self):
    #     example = bytearray.fromhex('060103' + ('09FF' + 255 * '61' + '092D' + 45 * '61') + '010568656c6c6f')
    #     expected = {
    #         6: bytearray(b'\x03'),
    #         9: bytearray(300 * b'a'),
    #         1: bytearray(b'hello')
    #     }
    #
    #     data = TLV.decode_bytearray(example)
    #     self.assertDictEqual(data, expected)
    #
    # def test_decode_bytearray_not_enough_data(self):
    #     example = bytearray.fromhex('060103' + '09FF' + 25 * '61')  # should have been 255 '61'
    #     self.assertRaises(TlvParseException, TLV.decode_bytearray, example)

    def test_decode_bytearray_to_list_not_enough_data(self):
        example = bytearray.fromhex("060103" + "09FF" + 25 * "61")  # should have been 255 '61'
        self.assertRaises(TlvParseException, TLV.decode_bytearray, example)

    def test_decode_bytes_to_list_not_enough_data(self):
        example = bytes(bytearray.fromhex("060103" + "09FF" + 25 * "61"))  # should have been 255 '61'
        self.assertRaises(TlvParseException, TLV.decode_bytes, example)

    def test_encode_list_key_error(self):
        example = [
            (
                -1,
                "hello",
            ),
        ]
        self.assertRaises(ValueError, TLV.encode_list, example)
        example = [
            (
                256,
                "hello",
            ),
        ]
        self.assertRaises(ValueError, TLV.encode_list, example)
        example = [
            (
                "test",
                "hello",
            ),
        ]
        self.assertRaises(ValueError, TLV.encode_list, example)

    def test_to_string_for_list(self):
        example = [
            (
                1,
                "hello",
            ),
        ]
        res = TLV.to_string(example)
        self.assertEqual(res, "[\n  1 (Identifier): (5 bytes/<class 'str'>) hello\n]\n")
        example = [
            (
                1,
                "hello",
            ),
            (
                2,
                "world",
            ),
        ]
        res = TLV.to_string(example)
        self.assertEqual(
            res,
            "[\n  1 (Identifier): (5 bytes/<class 'str'>) hello\n  2 (Salt): (5 bytes/<class 'str'>) world\n]\n",
        )

    def test_to_string_for_dict(self):
        example = {1: "hello"}
        res = TLV.to_string(example)
        self.assertEqual(res, "{\n  1 (Identifier): (5 bytes/<class 'str'>) hello\n}\n")
        example = {1: "hello", 2: "world"}
        res = TLV.to_string(example)
        self.assertEqual(
            res,
            "{\n  1 (Identifier): (5 bytes/<class 'str'>) hello\n  2 (Salt): (5 bytes/<class 'str'>) world\n}\n",
        )

    def test_to_string_for_dict_bytearray(self):
        example = {1: bytearray([0x42, 0x23])}
        res = TLV.to_string(example)
        self.assertEqual(res, "{\n  1 (Identifier): (2 bytes/<class 'bytearray'>) 0x4223\n}\n")

    def test_to_string_for_list_bytearray(self):
        example = [[1, bytearray([0x42, 0x23])]]
        res = TLV.to_string(example)
        self.assertEqual(res, "[\n  1 (Identifier): (2 bytes/<class 'bytearray'>) 0x4223\n]\n")

    def test_separator_list(self):
        val = [
            [TLV.kTLVType_State, TLV.M3],
            TLV.kTLVType_Separator_Pair,
            [TLV.kTLVType_State, TLV.M4],
        ]
        res = TLV.decode_bytearray(TLV.encode_list(val))
        self.assertEqual(val, res)

    def test_separator_list_error(self):
        val = [
            [TLV.kTLVType_State, TLV.M3],
            [TLV.kTLVType_Separator, "test"],
            [TLV.kTLVType_State, TLV.M4],
        ]
        self.assertRaises(ValueError, TLV.encode_list, val)

    def test_filter(self):
        example = bytes(bytearray.fromhex("060103" + "010203"))
        expected = [
            [6, bytearray(b"\x03")],
        ]

        data = TLV.decode_bytes(example, expected=[6])
        self.assertListEqual(data, expected)



from __future__ import annotations

from typing import Any
from unittest.mock import MagicMock

from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
import pytest

from aiohomekit.characteristic_cache import CharacteristicCacheMemory
from aiohomekit.controller.ble.controller import BleController

ADVERTISEMENT_DATA_DEFAULTS = {
    "local_name": "",
    "manufacturer_data": {},
    "service_data": {},
    "service_uuids": [],
    "rssi": -127,
    "platform_data": ((),),
    "tx_power": -127,
}

BLE_DEVICE_DEFAULTS = {
    "name": None,
    "rssi": -127,
    "details": None,
}


def generate_advertisement_data(**kwargs: Any) -> AdvertisementData:
    """Generate advertisement data with defaults."""
    new = kwargs.copy()
    for key, value in ADVERTISEMENT_DATA_DEFAULTS.items():
        new.setdefault(key, value)
    return AdvertisementData(**new)


def generate_ble_device(
    address: str | None = None,
    name: str | None = None,
    details: Any | None = None,
    rssi: int | None = None,
    **kwargs: Any,
) -> BLEDevice:
    """Generate a BLEDevice with defaults."""
    new = kwargs.copy()
    if address is not None:
        new["address"] = address
    if name is not None:
        new["name"] = name
    if details is not None:
        new["details"] = details
    if rssi is not None:
        new["rssi"] = rssi
    for key, value in BLE_DEVICE_DEFAULTS.items():
        new.setdefault(key, value)
    return BLEDevice(**new)


@pytest.fixture
def mock_bleak_scanner() -> MagicMock:
    return MagicMock()


@pytest.fixture
def ble_controller(mock_bleak_scanner: MagicMock) -> BleController:
    controller = BleController(CharacteristicCacheMemory(), mock_bleak_scanner)
    return controller


def test_discovery_with_none_name(mock_bleak_scanner: MagicMock, ble_controller: BleController) -> None:
    ble_device_with_short_name = generate_ble_device(name="Nam", address="00:00:00:00:00:00")
    ble_device_with_name = generate_ble_device(name="Name in Full", address="00:00:00:00:00:00")
    ble_device = generate_ble_device(
        name=None,
        address="00:00:00:00:00:00",
    )
    adv = generate_advertisement_data(
        local_name=None,
        manufacturer_data={76: b"\x061\x00\x80\xe7\x14j74\x06\x00\x9f!\x04\x02<\xb9\xeb\x0e"},
    )
    ble_controller._device_detected(ble_device, adv)
    assert "80:e7:14:6a:37:34" in ble_controller.discoveries
    ble_controller._device_detected(ble_device_with_short_name, adv)
    assert "80:e7:14:6a:37:34" in ble_controller.discoveries
    assert ble_controller.discoveries["80:e7:14:6a:37:34"].name == "Nam (00:00:00:00:00:00)"
    ble_controller._device_detected(ble_device, adv)
    assert ble_controller.discoveries["80:e7:14:6a:37:34"].name == "Nam (00:00:00:00:00:00)"
    ble_controller._device_detected(ble_device_with_name, adv)
    assert ble_controller.discoveries["80:e7:14:6a:37:34"].name == "Name in Full (00:00:00:00:00:00)"
    ble_controller._device_detected(ble_device_with_short_name, adv)
    assert ble_controller.discoveries["80:e7:14:6a:37:34"].name == "Name in Full (00:00:00:00:00:00)"



"""Test the AIO CLI variant."""

import json
from unittest import mock

import pytest

from aiohomekit.__main__ import main


async def test_help():
    with mock.patch("sys.stdout") as stdout:
        with pytest.raises(SystemExit):
            await main(["-h"])
    printed = stdout.write.call_args[0][0]

    assert printed.startswith("usage: ")
    assert "discover" in printed


async def test_get_accessories(pairing):
    with mock.patch("sys.stdout") as stdout:
        await main(["-f", "tests-pairing.json", "accessories", "-a", "alias"])
    printed = stdout.write.call_args_list[0][0][0]
    assert printed.startswith("1.1: >0000003E-0000-1000-8000-0026BB765291")

    with mock.patch("sys.stdout") as stdout:
        await main(["-f", "tests-pairing.json", "accessories", "-a", "alias", "-o", "json"])
    printed = stdout.write.call_args_list[0][0][0]
    accessories = json.loads(printed)
    assert accessories[0]["aid"] == 1
    assert accessories[0]["services"][0]["iid"] == 1
    assert accessories[0]["services"][0]["characteristics"][0]["iid"] == 2


async def test_get_characteristic(pairing):
    with mock.patch("sys.stdout") as stdout:
        await main(["-f", "tests-pairing.json", "get", "-a", "alias", "-c", "1.9"])
    printed = stdout.write.call_args_list[0][0][0]
    assert json.loads(printed) == {"1.9": {"value": False}}


async def test_put_characteristic(pairing):
    with mock.patch("sys.stdout"):
        await main(["-f", "tests-pairing.json", "put", "-a", "alias", "-c", "1.9", "true"])

    characteristics = await pairing.get_characteristics([(1, 9)])
    assert characteristics[(1, 9)] == {"value": True}


async def test_list_pairings(pairing):
    with mock.patch("sys.stdout") as stdout:
        await main(["-f", "tests-pairing.json", "list-pairings", "-a", "alias"])
    printed = "".join(write[0][0] for write in stdout.write.call_args_list)
    assert printed == (
        "Pairing Id: decc6fa3-de3e-41c9-adba-ef7409821bfc\n"
        "\tPublic Key: 0xd708df2fbf4a8779669f0ccd43f4962d6d49e4274f88b1292f822edc3bcf8ed8\n"
        "\tPermissions: 1 (admin)\n"
    )



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import asyncio
import pathlib
from asyncio.log import logger
from collections.abc import AsyncIterable
from contextlib import AsyncExitStack

from bleak import BleakScanner
from zeroconf.asyncio import AsyncZeroconf

from aiohomekit import hkjson
from aiohomekit.characteristic_cache import (
    CharacteristicCacheMemory,
    CharacteristicCacheType,
)
from aiohomekit.controller.abstract import AbstractDiscovery

from ..const import (
    BLE_TRANSPORT_SUPPORTED,
    COAP_TRANSPORT_SUPPORTED,
    IP_TRANSPORT_SUPPORTED,
)
from ..exceptions import (
    AccessoryNotFoundError,
    ConfigLoadingError,
    ConfigSavingError,
    TransportNotSupportedError,
)
from .abstract import AbstractController, AbstractPairing, TransportType


class Controller(AbstractController):
    """
    This class represents a HomeKit controller (normally your iPhone or iPad).
    """

    pairings: dict[str, AbstractPairing]

    def __init__(
        self,
        async_zeroconf_instance: AsyncZeroconf | None = None,
        char_cache: CharacteristicCacheType | None = None,
        bleak_scanner_instance: BleakScanner | None = None,
    ) -> None:
        """
        Initialize an empty controller. Use 'load_data()' to load the pairing data.

        :param ble_adapter: the bluetooth adapter to be used (defaults to hci0)
        """
        super().__init__(char_cache=char_cache or CharacteristicCacheMemory())

        self._async_zeroconf_instance = async_zeroconf_instance
        self._bleak_scanner_instance = bleak_scanner_instance

        self.transports: dict[TransportType, AbstractController] = {}
        self._tasks = AsyncExitStack()

    async def _async_register_backend(self, controller: AbstractController) -> None:
        self.transports[controller.transport_type] = await self._tasks.enter_async_context(controller)

    async def async_start(self) -> None:
        if IP_TRANSPORT_SUPPORTED or self._async_zeroconf_instance:
            from .ip.controller import (
                IpController,  # pylint: disable=import-outside-toplevel
            )

            await self._async_register_backend(
                IpController(
                    char_cache=self._char_cache,
                    zeroconf_instance=self._async_zeroconf_instance,
                )
            )

        if COAP_TRANSPORT_SUPPORTED:
            from .coap.controller import (
                CoAPController,  # pylint: disable=import-outside-toplevel
            )

            await self._async_register_backend(
                CoAPController(
                    char_cache=self._char_cache,
                    zeroconf_instance=self._async_zeroconf_instance,
                )
            )

        if BLE_TRANSPORT_SUPPORTED or self._bleak_scanner_instance:
            from .ble.controller import (
                BleController,  # pylint: disable=import-outside-toplevel
            )

            await self._async_register_backend(
                BleController(
                    char_cache=self._char_cache,
                    bleak_scanner_instance=self._bleak_scanner_instance,
                )
            )

    async def async_stop(self) -> None:
        await self._tasks.aclose()

    async def async_find(self, device_id: str, timeout: float = 30.0) -> AbstractDiscovery:
        pending = []
        for transport in self.transports.values():
            pending.append(asyncio.create_task(transport.async_find(device_id, timeout)))

        try:
            while pending:
                done, pending = await asyncio.wait(pending, return_when=asyncio.FIRST_COMPLETED)
                for result in done:
                    try:
                        return result.result()
                    except AccessoryNotFoundError:
                        continue
        finally:
            [task.cancel() for task in pending]
            for task in pending:
                try:
                    await task
                except asyncio.CancelledError:
                    pass

        raise AccessoryNotFoundError(f"Accessory with device id {device_id} not found")

    async def async_reachable(self, device_id: str, timeout=10) -> bool:
        raise NotImplementedError

    async def async_discover(self, timeout=10) -> AsyncIterable[AbstractDiscovery]:
        for transport in self.transports.values():
            async for device in transport.async_discover():
                yield device

    def load_pairing(self, alias: str, pairing_data: dict[str, str]) -> AbstractPairing:
        """
        Loads a pairing instance from a pairing data dict.
        """
        if "Connection" not in pairing_data:
            pairing_data["Connection"] = "IP"

        for transport in self.transports.values():
            if pairing := transport.load_pairing(alias, pairing_data):
                self.pairings[pairing_data["AccessoryPairingID"].lower()] = pairing
                self.aliases[alias] = pairing
                return pairing

        raise TransportNotSupportedError(pairing_data["Connection"])

    def load_data(self, filename: str) -> None:
        """
        Loads the pairing data of the controller from a file.

        :param filename: the file name of the pairing data
        :raises ConfigLoadingError: if the config could not be loaded. The reason is given in the message.
        """
        try:
            with open(filename, encoding="utf-8") as input_fp:
                data = hkjson.loads(input_fp.read())
                for pairing_id in data:
                    try:
                        self.load_pairing(pairing_id, data[pairing_id])
                    except TransportNotSupportedError as e:
                        logger.error("Skipped pairing: %s", e)
        except PermissionError:
            raise ConfigLoadingError(f'Could not open "{filename}" due to missing permissions')
        except hkjson.JSON_DECODE_EXCEPTIONS as e:
            raise ConfigLoadingError(f'Cannot parse "{filename}" as JSON file') from e
        except FileNotFoundError:
            pass

    def save_data(self, filename: str) -> None:
        """
        Saves the pairing data of the controller to a file.

        :param filename: the file name of the pairing data
        :raises ConfigSavingError: if the config could not be saved. The reason is given in the message.
        """
        data = {}
        for alias in self.aliases:
            data[alias] = self.aliases[alias].pairing_data

        path = pathlib.Path(filename)

        if not path.parent.exists():
            path.parent.mkdir(parents=True, exist_ok=True)

        try:
            with open(filename, mode="w", encoding="utf-8") as output_fp:
                output_fp.write(hkjson.dumps_indented(data))
        except PermissionError:
            raise ConfigSavingError(f'Could not write "{filename}" due to missing permissions')
        except FileNotFoundError:
            raise ConfigSavingError(f'Could not write "{filename}" because it (or the folder) does not exist')

    async def remove_pairing(self, alias: str) -> None:
        """
        Remove a pairing between the controller and the accessory. The pairing data is delete on both ends, on the
        accessory and the controller.

        Important: no automatic saving of the pairing data is performed. If you don't do this, the accessory seems still
            to be paired on the next start of the application.

        :param alias: the controller's alias for the accessory
        :raises AuthenticationError: if the controller isn't authenticated to the accessory.
        :raises AccessoryNotFoundError: if the device can not be found via zeroconf
        :raises UnknownError: on unknown errors
        """
        if alias not in self.aliases:
            raise AccessoryNotFoundError(f'Alias "{alias}" is not found.')

        pairing = self.aliases[alias]

        try:
            try:
                # Remove the pairing from the controller first
                # so that it stops getting updates that might
                # trigger a disconnected event poll.
                self.aliases.pop(alias, None)
                pairing.controller.aliases.pop(alias, None)

                self.pairings.pop(pairing.id, None)
                pairing.controller.pairings.pop(pairing.id, None)

                primary_pairing_id = pairing.pairing_data["iOSPairingId"]

                await pairing.remove_pairing(primary_pairing_id)
            finally:
                await pairing.shutdown()
        finally:
            # Outer finally block to ensure that the pairing is removed
            # from the controller even if the shutdown fails.
            self._char_cache.async_delete_map(pairing.id)



"""Tests for format_characteristic_list function in IP pairing."""

from aiohomekit.controller.ip.pairing import format_characteristic_list


def test_format_characteristic_list_normal_response() -> None:
    """Test normal response with characteristic values."""
    response = {
        "characteristics": [
            {"aid": 1, "iid": 2, "value": 23.5},
            {"aid": 1, "iid": 3, "value": "test"},
            {"aid": 2, "iid": 4, "value": True},
        ]
    }
    result = format_characteristic_list(response)

    assert len(result) == 3
    assert result[(1, 2)] == {"value": 23.5}
    assert result[(1, 3)] == {"value": "test"}
    assert result[(2, 4)] == {"value": True}


def test_format_characteristic_list_success_status_removed() -> None:
    """Test that status=0 is removed from successful responses."""
    response = {
        "characteristics": [
            {"aid": 1, "iid": 2, "value": 23.5, "status": 0},
            {"aid": 1, "iid": 3, "value": "test"},
        ]
    }
    result = format_characteristic_list(response)

    assert result[(1, 2)] == {"value": 23.5}  # status=0 should be removed
    assert "status" not in result[(1, 2)]
    assert result[(1, 3)] == {"value": "test"}


def test_format_characteristic_list_individual_errors() -> None:
    """Test response with individual characteristic errors."""
    response = {
        "characteristics": [
            {"aid": 1, "iid": 2, "value": 23.5},
            {"aid": 1, "iid": 3, "status": -70402},
            {"aid": 2, "iid": 4, "status": -70404},
        ]
    }
    result = format_characteristic_list(response)

    assert result[(1, 2)] == {"value": 23.5}
    assert result[(1, 3)]["status"] == -70402
    assert "description" in result[(1, 3)]
    assert result[(2, 4)]["status"] == -70404
    assert "description" in result[(2, 4)]


def test_format_characteristic_list_global_error_no_requested() -> None:
    """Test global error response without knowing what was requested."""
    response = {"status": -70407}
    result = format_characteristic_list(response)

    assert result == {}


def test_format_characteristic_list_global_error_with_requested() -> None:
    """Test global error response with requested characteristics."""
    response = {"status": -70407}
    requested = {(1, 2), (1, 3), (2, 4)}

    result = format_characteristic_list(response, requested)

    assert len(result) == 3
    for aid, iid in requested:
        assert (aid, iid) in result
        assert result[(aid, iid)]["status"] == -70407
        assert "description" in result[(aid, iid)]
        assert "Out of resources" in result[(aid, iid)]["description"]


def test_format_characteristic_list_various_global_errors() -> None:
    """Test different global error status codes."""
    test_cases = [
        (-70401, "Request denied due to insufficient privileges"),
        (-70402, "Unable to communicate"),
        (-70403, "Resource is busy"),
        (-70408, "Operation timed out"),
        (-70409, "Resource does not exist"),
    ]

    for status, expected_desc_part in test_cases:
        response = {"status": status}
        requested = {(1, 2)}

        result = format_characteristic_list(response, requested)

        assert result[(1, 2)]["status"] == status
        assert expected_desc_part in result[(1, 2)]["description"]


def test_format_characteristic_list_empty_array() -> None:
    """Test response with empty characteristics array."""
    response = {"characteristics": []}
    result = format_characteristic_list(response)

    assert result == {}


def test_format_characteristic_list_mixed_success_failure() -> None:
    """Test response with mix of successful and failed characteristics."""
    response = {
        "characteristics": [
            {"aid": 1, "iid": 2, "value": 100, "status": 0},
            {"aid": 1, "iid": 3, "status": -70410},
            {"aid": 2, "iid": 4, "value": "OK"},
        ]
    }
    result = format_characteristic_list(response)

    assert result[(1, 2)] == {"value": 100}
    assert "status" not in result[(1, 2)]
    assert result[(1, 3)]["status"] == -70410
    assert "value" not in result[(1, 3)]
    assert result[(2, 4)] == {"value": "OK"}


def test_format_characteristic_list_large_request_global_error() -> None:
    """Test global error with large number of requested characteristics (like the real issue)."""
    # Simulate the actual scenario from the logs
    response = {"status": -70407}

    # Create a large set of requested characteristics similar to the actual error
    requested = set()
    for aid in range(1, 8):  # 7 accessories
        for iid in range(3, 20):  # Multiple characteristics each
            requested.add((aid, iid))

    result = format_characteristic_list(response, requested)

    assert len(result) == len(requested)
    for char_id in requested:
        assert char_id in result
        assert result[char_id]["status"] == -70407
        assert "Out of resources" in result[char_id]["description"]


def test_format_characteristic_list_status_zero_only() -> None:
    """Test response where all characteristics have status=0 (success)."""
    response = {
        "characteristics": [
            {"aid": 1, "iid": 2, "status": 0},
            {"aid": 1, "iid": 3, "status": 0},
        ]
    }
    result = format_characteristic_list(response)

    # Status=0 should be removed, leaving empty dicts
    assert result[(1, 2)] == {}
    assert result[(1, 3)] == {}


def test_format_characteristic_list_malformed_characteristics() -> None:
    """Test handling of malformed characteristics missing aid or iid."""
    response = {
        "characteristics": [
            {"aid": 1, "iid": 2, "value": 23.5},
            {"iid": 3, "value": "missing aid"},  # Missing aid
            {"aid": 2, "value": "missing iid"},  # Missing iid
            {"value": "missing both"},  # Missing both aid and iid
            {"aid": 1, "iid": 4, "value": 100},
        ]
    }
    result = format_characteristic_list(response)

    # Should only have the valid characteristics
    assert len(result) == 2
    assert result[(1, 2)] == {"value": 23.5}
    assert result[(1, 4)] == {"value": 100}
    # Malformed ones should be skipped


def test_format_characteristic_list_global_error_with_partial_data() -> None:
    """Test global error with some characteristics present."""
    # Device returns global error but also includes some successful characteristics
    response = {
        "status": -70407,
        "characteristics": [
            {"aid": 1, "iid": 2, "value": 23.5},
            {"aid": 1, "iid": 3, "status": -70402},
        ],
    }
    requested = {(1, 2), (1, 3), (2, 4), (2, 5)}

    result = format_characteristic_list(response, requested)

    # Should have all 4 requested characteristics
    assert len(result) == 4

    # The ones in the response should use their actual data
    assert result[(1, 2)] == {"value": 23.5}
    assert result[(1, 3)]["status"] == -70402
    assert "Unable to communicate" in result[(1, 3)]["description"]

    # The missing ones should get the global error
    assert result[(2, 4)]["status"] == -70407
    assert "Out of resources" in result[(2, 4)]["description"]
    assert result[(2, 5)]["status"] == -70407
    assert "Out of resources" in result[(2, 5)]["description"]


def test_format_characteristic_list_unknown_status_codes() -> None:
    """Test handling of unknown/undefined status codes including -6 from Nanoleaf."""
    # Test status -6 (from Nanoleaf devices)
    response = {"status": -6}
    requested = {(1, 2), (1, 3)}

    result = format_characteristic_list(response, requested)

    assert len(result) == 2
    assert result[(1, 2)]["status"] == -6
    assert result[(1, 2)]["description"] == "Unknown error code: -6"
    assert result[(1, 3)]["status"] == -6
    assert result[(1, 3)]["description"] == "Unknown error code: -6"

    # Test another unknown error
    response = {"status": -99999}
    requested = {(1, 2)}

    result = format_characteristic_list(response, requested)

    assert result[(1, 2)]["status"] == -99999
    assert result[(1, 2)]["description"] == "Unknown error code: -99999"

    # Test individual unknown error
    response = {
        "characteristics": [
            {"aid": 1, "iid": 2, "value": 23.5},
            {"aid": 1, "iid": 3, "status": -88888},
        ]
    }

    result = format_characteristic_list(response)

    assert result[(1, 2)] == {"value": 23.5}
    assert result[(1, 3)]["status"] == -88888
    assert result[(1, 3)]["description"] == "Unknown error code: -88888"


def test_format_characteristic_list_real_thermostat_scenario() -> None:
    """Test the exact scenario from the Resideo T9/T10 thermostat error."""
    # This is the exact response from the device when overwhelmed
    response = {"status": -70407}

    # Sample of characteristics requested (actual request had 103)
    requested = {
        (7, 81),
        (4, 9),
        (2, 39),
        (5, 65),
        (5, 10),
        (2, 130),
        (5, 37),
        (3, 49),
        (6, 66),
        (7, 65),
    }

    result = format_characteristic_list(response, requested)

    assert len(result) == 10
    for char_id in requested:
        assert char_id in result
        assert result[char_id]["status"] == -70407
        assert result[char_id]["description"] == "Out of resources to process request."


def test_format_characteristic_list_boolean_in_response() -> None:
    """Test handling of boolean values in characteristics array (from TaHoma covers issue #465)."""
    # This is the exact response that caused the issue in the TaHoma V2 covers
    # When setting tilt positions, the device sometimes returns boolean values
    # mixed with error dictionaries
    response = {
        "characteristics": [
            True,  # Boolean instead of dict
            {"iid": 1400123, "aid": 490605949956, "status": -70402},
            True,  # Another boolean
        ]
    }
    result = format_characteristic_list(response)

    # Should only have the valid characteristic with error
    assert len(result) == 1
    assert result[(490605949956, 1400123)]["status"] == -70402
    assert "Unable to communicate" in result[(490605949956, 1400123)]["description"]



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import base64

from aiohomekit.model import Accessories
from aiohomekit.model.characteristics import CharacteristicsTypes
from aiohomekit.model.characteristics.const import (
    AudioCodecValues,
    BitRateValues,
    PacketizationModeValues,
    ProfileIDValues,
    ProfileSupportLevelValues,
    SampleRateValues,
    SRTPCryptoSuiteValues,
    StreamingStatusValues,
    ThreadStatus,
    VideoCodecTypeValues,
)
from aiohomekit.model.characteristics.structs import (
    AudioCodecConfiguration,
    AudioCodecParameters,
    SupportedAudioStreamConfiguration,
    SupportedRTPConfiguration,
    SupportedVideoStreamConfiguration,
    VideoAttrs,
    VideoCodecParameters,
    VideoConfigConfiguration,
)
from aiohomekit.model.services import ServicesTypes
from aiohomekit.protocol.statuscodes import HapStatusCode


def test_hue_bridge():
    a = Accessories.from_file("tests/fixtures/hue_bridge.json").aid(6623462389072572)
    assert a.name == "Hue dimmer switch"
    assert a.model == "RWL021"
    assert a.manufacturer == "Philips"
    assert a.serial_number == "6623462389072572"
    assert a.firmware_revision == "45.1.17846"

    service = a.services.first(service_type=ServicesTypes.ACCESSORY_INFORMATION)

    char = next(iter(service.characteristics))
    assert char.iid == 37
    assert char.perms == ["pr"]
    assert char.format == "string"
    assert char.value == "Hue dimmer switch"
    assert char.description == "Name"
    assert char.maxLen == 64

    assert service.has(char.type)


def test_idevices_switch():
    a = Accessories.from_file("tests/fixtures/idevices_switch.json").aid(1)
    assert a.name == "iDevices Switch"
    assert a.model == "IDEV0001"
    assert a.manufacturer == "iDevices LLC"
    assert a.serial_number == "00080685"
    assert a.firmware_revision == "1.2.1"

    service = a.services.first(service_type=ServicesTypes.ACCESSORY_INFORMATION)

    char = next(iter(service.characteristics))
    assert char.iid == 2
    assert char.perms == ["pr"]
    assert char.format == "string"
    assert char.value == "iDevices Switch"
    assert char.description == "Name"
    assert char.maxLen == 64

    assert service.has(char.type)


def test_linked_services():
    a = Accessories.from_file("tests/fixtures/hue_bridge.json").aid(6623462389072572)

    service = a.services.first(service_type=ServicesTypes.STATELESS_PROGRAMMABLE_SWITCH)
    assert len(service.linked) > 0
    assert service.linked[0].type == "000000CC-0000-1000-8000-0026BB765291"


def test_get_by_name():
    name = "Hue dimmer switch button 3"
    a = Accessories.from_file("tests/fixtures/hue_bridge.json").aid(6623462389072572)

    service = a.services.first(
        service_type=ServicesTypes.STATELESS_PROGRAMMABLE_SWITCH,
        characteristics={CharacteristicsTypes.NAME: name},
    )

    assert service[CharacteristicsTypes.NAME].value == name


def test_get_by_characteristic_types():
    name = "Hue dimmer switch button 3"

    a = Accessories.from_file("tests/fixtures/hue_bridge.json").aid(6623462389072572)

    service = a.services.first(
        service_type=ServicesTypes.STATELESS_PROGRAMMABLE_SWITCH,
        characteristics={CharacteristicsTypes.NAME: name},
    )

    char = service.characteristics.first(char_types=[CharacteristicsTypes.NAME])

    assert char.value == name


def test_get_by_iid():
    name = "Hue dimmer switch button 3"

    bridge = Accessories.from_file("tests/fixtures/hue_bridge.json")
    assert bridge.has_aid(6623462389072572) is True
    assert bridge.aid_or_none(6623462389072572) is not None
    assert bridge.has_aid(9999) is False
    assert bridge.aid_or_none(9999) is None

    accessory = bridge.aid(6623462389072572)

    char = accessory.characteristics.iid(588410716196)
    assert char.value == name

    service = accessory.services.first(
        service_type=ServicesTypes.STATELESS_PROGRAMMABLE_SWITCH,
        characteristics={CharacteristicsTypes.NAME: name},
    )
    assert service.get_char_by_iid(588410716196) is char
    assert accessory.services.iid_or_none(service.iid) is service
    assert accessory.services.iid_or_none(9999) is None


def test_get_by_vendor_characteristic_types():
    spray_level = 5

    a = Accessories.from_file("tests/fixtures/vocolinc_flowerbud.json").aid(1)

    service = a.services.first(
        service_type=ServicesTypes.HUMIDIFIER_DEHUMIDIFIER,
    )

    found = service.has(CharacteristicsTypes.VENDOR_VOCOLINC_HUMIDIFIER_SPRAY_LEVEL)
    assert found is True

    char = service.characteristics.first(
        char_types=[CharacteristicsTypes.VENDOR_VOCOLINC_HUMIDIFIER_SPRAY_LEVEL]
    )
    assert char.value == spray_level


def test_get_by_service_type_is_normalized():
    """Test we can get a service by a normalized service type."""
    a = Accessories.from_file("tests/fixtures/vocolinc_flowerbud.json").aid(1)

    service = a.services.first(
        service_type=ServicesTypes.HUMIDIFIER_DEHUMIDIFIER.lower(),
    )

    found = service.has(CharacteristicsTypes.VENDOR_VOCOLINC_HUMIDIFIER_SPRAY_LEVEL)
    assert found is True


def test_get_by_linked():
    name = "Hue dimmer switch button 3"
    a = Accessories.from_file("tests/fixtures/hue_bridge.json").aid(6623462389072572)

    switch = a.services.first(
        service_type=ServicesTypes.STATELESS_PROGRAMMABLE_SWITCH,
        characteristics={CharacteristicsTypes.NAME: name},
    )

    service_label = a.services.first(parent_service=switch)
    assert service_label[CharacteristicsTypes.SERVICE_LABEL_NAMESPACE].value == 1

    switch = a.services.first(
        service_type=ServicesTypes.STATELESS_PROGRAMMABLE_SWITCH,
        characteristics={CharacteristicsTypes.NAME: name},
        child_service=service_label,
    )

    assert switch[CharacteristicsTypes.NAME].value == "Hue dimmer switch button 3"


def test_order_by():
    a = Accessories.from_file("tests/fixtures/hue_bridge.json").aid(6623462389072572)

    buttons = a.services.filter(
        service_type=ServicesTypes.STATELESS_PROGRAMMABLE_SWITCH,
        order_by=(CharacteristicsTypes.SERVICE_LABEL_INDEX, CharacteristicsTypes.NAME),
    )

    assert buttons[0].value(CharacteristicsTypes.SERVICE_LABEL_INDEX) == 1
    assert buttons[1].value(CharacteristicsTypes.SERVICE_LABEL_INDEX) == 2
    assert buttons[2].value(CharacteristicsTypes.SERVICE_LABEL_INDEX) == 3
    assert buttons[3].value(CharacteristicsTypes.SERVICE_LABEL_INDEX) == 4


def test_process_changes():
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")

    on_char = accessories.aid(1).characteristics.iid(8)
    assert on_char.value is False

    changed = accessories.process_changes({(1, 8): {"value": True}})
    assert changed == {(1, 8)}

    changed = accessories.process_changes({(1, 8): {"value": True}})
    assert changed == set()

    assert on_char.value is True


def test_process_changes_error():
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")

    on_char = accessories.aid(1).characteristics.iid(8)
    assert on_char.value is False
    assert on_char.status == HapStatusCode.SUCCESS

    changed = accessories.process_changes({(1, 8): {"status": HapStatusCode.UNABLE_TO_COMMUNICATE.value}})

    assert on_char.value is False
    assert on_char.status == HapStatusCode.UNABLE_TO_COMMUNICATE
    assert changed == {(1, 8)}

    changed = accessories.process_changes({(1, 8): {"value": True}})
    assert changed == {(1, 8)}
    assert on_char.value is True
    assert on_char.status == HapStatusCode.SUCCESS


def test_process_changes_availability():
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")

    on_char = accessories.aid(1).characteristics.iid(8)

    assert on_char.available is True
    assert on_char.service.available is True
    assert on_char.service.accessory.available is True

    changed = accessories.process_changes({(1, 8): {"status": HapStatusCode.UNABLE_TO_COMMUNICATE.value}})
    assert changed == {(1, 8)}

    assert on_char.available is False
    assert on_char.service.available is False
    assert on_char.service.accessory.available is False

    changed = accessories.process_changes({(1, 8): {"value": True}})
    assert changed == {(1, 8)}
    assert on_char.available is True
    assert on_char.service.available is True
    assert on_char.service.accessory.available is True


def test_valid_vals_preserved():
    a = Accessories.from_file("tests/fixtures/aqara_gateway.json").aid(1)
    char = a.characteristics.iid(66307)
    assert char.valid_values == [1, 3, 4]


def test_build_update():
    name = "Hue dimmer switch button 3"

    a = Accessories.from_file("tests/fixtures/hue_bridge.json").aid(6623462389072572)

    service = a.services.first(
        service_type=ServicesTypes.STATELESS_PROGRAMMABLE_SWITCH,
        characteristics={CharacteristicsTypes.NAME: name},
    )

    payload = service.build_update({CharacteristicsTypes.NAME: "Fred"})

    assert payload == [(6623462389072572, 588410716196, "Fred")]


def test_build_update_minStep_clamping_lennox():
    a = Accessories.from_file("tests/fixtures/lennox_e30.json").aid(1)
    service = a.services.iid(100)

    assertions = [
        (27.23, 27.0),
        (27.6, 27.5),
        (27.26, 27.5),
        (27.9, 28.0),
    ]

    for left, right in assertions:
        payload = service.build_update({CharacteristicsTypes.TEMPERATURE_TARGET: left})
        assert payload == [(1, 104, right)]


def test_build_update_minStep_clamping_ecobee():
    a = Accessories.from_file("tests/fixtures/ecobee3.json").aid(1)
    service = a.services.iid(16)

    assertions = [
        (27.23, 27.2),
        (27.6, 27.6),
        (27.26, 27.3),
        (27.9, 27.9),
        (27.95, 28.0),
    ]

    for left, right in assertions:
        payload = service.build_update({CharacteristicsTypes.TEMPERATURE_TARGET: left})
        assert payload == [(1, 20, right)]


def test_build_update_minStep_clamping_synthetic():
    a = Accessories.from_file("tests/fixtures/synthetic_float_minstep.json")

    assertions = [
        # minStep 1
        (1, 27.2, 27.5),
        (1, 27.6, 27.5),
        (1, 27.9, 27.5),
        # minStep 2
        (2, 27.2, 26.5),
        (2, 28.2, 28.5),
        (2, 27.7, 28.5),
        # minStep 5
        (3, 27.2, 29.5),
        (3, 25.0, 24.5),
        (3, 28.3, 29.5),
        # no minStep
        (4, 27.2, 27.2),
        (4, 27.3, 27.3),
        (4, 27.7, 27.7),
        # minStep 1, no offset
        (5, 27.2, 27.0),
        (5, 27.6, 28.0),
        (5, 27.9, 28.0),
        # minStep 2, no offset
        (6, 27.2, 28.0),
        (6, 28.2, 28.0),
        (6, 27.7, 28.0),
        # minStep 5, no offset
        (7, 27.2, 25.0),
        (7, 25.0, 25.0),
        (7, 28.3, 30.0),
    ]

    for aid, left, right in assertions:
        service = a.aid(aid).services.iid(100)
        payload = service.build_update({CharacteristicsTypes.TEMPERATURE_TARGET: left})
        assert payload == [(aid, 104, right)]


def test_build_update_minStep_clamping_synthetic_int():
    a = Accessories.from_file("tests/fixtures/synthetic_float_minstep.json")

    assertions = [
        # minStep 1, int
        (8, 27.0, 27),
        (8, 27.5, 28),
        (8, 28.0, 28),
        (8, 28.5, 29),
        (8, 29.0, 29),
        (8, 29.5, 30),
        (8, 27.2, 27),
        (8, 27.6, 28),
        (8, 27.9, 28),
    ]

    for aid, left, right in assertions:
        service = a.aid(aid).services.iid(100)
        payload = service.build_update({CharacteristicsTypes.TEMPERATURE_TARGET: left})
        assert payload == [(aid, 104, right)]
        assert isinstance(payload[0][2], int)


def test_tlv8_struct():
    a = Accessories.from_file("tests/fixtures/home_assistant_bridge_camera.json")
    service = a.aid(2018094878).services.iid(11)

    streaming_status = service.value(CharacteristicsTypes.STREAMING_STATUS)
    assert streaming_status.status == StreamingStatusValues.AVAILABLE

    video_stream_config = service.value(CharacteristicsTypes.SUPPORTED_VIDEO_STREAM_CONFIGURATION)

    assert video_stream_config == SupportedVideoStreamConfiguration(
        config=[
            VideoConfigConfiguration(
                codec_type=VideoCodecTypeValues.H264,
                codec_params=[
                    VideoCodecParameters(
                        profile_id=ProfileIDValues.HIGH_PROFILE,
                        level=ProfileSupportLevelValues.FOUR,
                        packetization_mode=PacketizationModeValues.NON_INTERLEAVED_MODE,
                        cvo_enabled=None,
                        cvo_id=None,
                    )
                ],
                video_attrs=[VideoAttrs(width=1920, height=1080, fps=30)],
            )
        ]
    )

    audio_stream_config = service.value(CharacteristicsTypes.SUPPORTED_AUDIO_CONFIGURATION)

    assert audio_stream_config == SupportedAudioStreamConfiguration(
        config=[
            AudioCodecConfiguration(
                codec=AudioCodecValues.OPUS,
                parameters=[
                    AudioCodecParameters(
                        audio_channels=1,
                        bit_rate=BitRateValues.VARIABLE,
                        sample_rate=SampleRateValues.SIXTEEN_KHZ,
                        rtp_time=None,
                    )
                ],
            )
        ],
        comfort_noise=0,
    )

    rtp_config = service.value(CharacteristicsTypes.SUPPORTED_RTP_CONFIGURATION)

    assert rtp_config == [
        SupportedRTPConfiguration(
            srtp_crypto_suite=SRTPCryptoSuiteValues.AES_CM_128_HMAC_SHA1_80,
        ),
    ]


def test_tlv8_struct_bare_array():
    a = Accessories.from_file("tests/fixtures/camera.json")
    service = a.aid(1).services.iid(16)

    rtp_config = service.value(CharacteristicsTypes.SUPPORTED_RTP_CONFIGURATION)

    assert rtp_config == [
        SupportedRTPConfiguration(
            srtp_crypto_suite=SRTPCryptoSuiteValues.DISABLED,
        ),
        SupportedRTPConfiguration(
            srtp_crypto_suite=SRTPCryptoSuiteValues.AES_CM_128_HMAC_SHA1_80,
        ),
    ]


def test_tlv8_struct_re_encode():
    a = Accessories.from_file("tests/fixtures/camera.json")
    service = a.aid(1).services.iid(16)

    video_stream_config = service.value(CharacteristicsTypes.SUPPORTED_VIDEO_STREAM_CONFIGURATION)

    raw = base64.b64decode(
        "AcUBAQACHQEBAAAAAQEBAAABAQICAQAAAAIBAQAAAgECAwEAAwsBAoAHAgI4BAMBHgAAAwsBAgAFAgL"
        "AAwMBHgAAAwsBAgAFAgLQAgMBHgAAAwsBAgAEAgIAAwMBHgAAAwsBAoACAgLgAQMBHgAAAwsBAoACAg"
        "JoAQMBHgAAAwsBAuABAgJoAQMBHgAAAwsBAuABAgIOAQMBHgAAAwsBAkABAgLwAAMBHgAAAwsBAkABA"
        "gLwAAMBDwAAAwsBAkABAgK0AAMBHg=="
    )

    assert raw == video_stream_config.encode()


def test_set_value():
    """
    At the moment a bunch of tests in Home Assistant rely on Char.set_value
    """
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")

    on_char = accessories.aid(1).characteristics.iid(8)
    assert on_char.value is False

    on_char.value = True

    assert on_char.value is True


def test_enum_decode():
    accessories = Accessories.from_file("tests/fixtures/nanoleaf_bulb.json")
    status = accessories.aid(1).characteristics.iid(117)
    assert status.value == ThreadStatus.ROUTER


def test_enum_encode():
    accessories = Accessories.from_file("tests/fixtures/nanoleaf_bulb.json")
    status = accessories.aid(1).characteristics.iid(117)
    status.value = ThreadStatus.DISABLED
    assert status.to_accessory_and_service_list()["value"] == 1


def test_needs_polling():
    accessories = Accessories.from_file("tests/fixtures/eve_energy.json")
    assert any(a.needs_polling for a in accessories) is True

    accessories = Accessories.from_file("tests/fixtures/nanoleaf_bulb.json")
    assert any(a.needs_polling for a in accessories) is False



from unittest import mock

from aiohomekit.model import Accessories, Accessory
from aiohomekit.model.characteristics import CharacteristicsTypes
from aiohomekit.model.services import ServicesTypes
from aiohomekit.protocol.statuscodes import HapStatusCode
from aiohomekit.testing import FakeController


async def test_pairing():
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")
    controller = FakeController()
    device = controller.add_device(accessories)

    discovery = await controller.async_find(device.description.id)
    finish_pairing = await discovery.async_start_pairing("alias")
    pairing = await finish_pairing("111-22-333")

    chars_and_services = await pairing.list_accessories_and_characteristics()
    assert isinstance(chars_and_services, list)


async def test_get_and_set():
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")
    controller = FakeController()
    device = controller.add_device(accessories)

    discovery = await controller.async_find(device.description.id)
    finish_pairing = await discovery.async_start_pairing("alias")
    pairing = await finish_pairing("111-22-333")

    pairing.dispatcher_connect(accessories.process_changes)

    chars = await pairing.get_characteristics([(1, 10)])
    assert chars == {(1, 10): {"value": 0}}

    chars = await pairing.put_characteristics([(1, 10, 1)])
    assert chars == {}

    chars = await pairing.get_characteristics([(1, 10)])
    assert chars == {(1, 10): {"value": 1}}


async def test_get_failure():
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")

    char = accessories.aid(1).characteristics.iid(10)
    char.status = HapStatusCode.UNABLE_TO_COMMUNICATE

    controller = FakeController()
    device = controller.add_device(accessories)

    discovery = await controller.async_find(device.description.id)
    finish_pairing = await discovery.async_start_pairing("alias")
    pairing = await finish_pairing("111-22-333")

    chars = await pairing.get_characteristics([(1, 10)])
    assert chars == {(1, 10): {"status": -70402}}


async def test_put_failure():
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")

    char = accessories.aid(1).characteristics.iid(10)
    char.status = HapStatusCode.UNABLE_TO_COMMUNICATE

    controller = FakeController()
    device = controller.add_device(accessories)

    discovery = await controller.async_find(device.description.id)
    finish_pairing = await discovery.async_start_pairing("alias")
    pairing = await finish_pairing("111-22-333")

    chars = await pairing.put_characteristics([(1, 10, 1)])
    assert chars == {(1, 10): {"status": -70402}}


async def test_update_named_service_events():
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")
    controller = FakeController()
    pairing = await controller.add_paired_device(accessories, "alias")

    await pairing.subscribe([(1, 8)])
    pairing.dispatcher_connect(accessories.process_changes)

    # Simulate that the state was changed on the device itself.
    pairing.testing.update_named_service("Light Strip", {CharacteristicsTypes.ON: True})

    assert accessories.aid(1).characteristics.iid(8).value == 1


async def test_update_named_service_events_manual_accessory(id_factory):
    accessories = Accessories()
    accessory = Accessory.create_with_info(
        id_factory(),
        name="TestLight",
        manufacturer="Test Mfr",
        model="Test Bulb",
        serial_number="1234",
        firmware_revision="1.1",
    )
    service = accessory.add_service(ServicesTypes.LIGHTBULB, name="Light Strip")
    on_char = service.add_char(CharacteristicsTypes.ON)
    accessories.add_accessory(accessory)

    controller = FakeController()
    pairing = await controller.add_paired_device(accessories, "alias")

    callback = mock.Mock()
    await pairing.subscribe([(accessory.aid, on_char.iid)])
    pairing.dispatcher_connect(callback)

    # Simulate that the state was changed on the device itself.
    pairing.testing.update_named_service("Light Strip", {CharacteristicsTypes.ON: True})

    assert callback.call_args_list == [mock.call({(accessory.aid, on_char.iid): {"value": 1}})]


async def test_update_named_service_events_manual_accessory_auto_requires(id_factory):
    accessories = Accessories()
    accessory = Accessory.create_with_info(
        id_factory(),
        name="TestLight",
        manufacturer="Test Mfr",
        model="Test Bulb",
        serial_number="1234",
        firmware_revision="1.1",
    )
    service = accessory.add_service(ServicesTypes.LIGHTBULB, name="Light Strip", add_required=True)
    on_char = service[CharacteristicsTypes.ON]
    accessories.add_accessory(accessory)

    controller = FakeController()
    pairing = await controller.add_paired_device(accessories, "alias")

    callback = mock.Mock()
    await pairing.subscribe([(accessory.aid, on_char.iid)])
    pairing.dispatcher_connect(callback)

    # Simulate that the state was changed on the device itself.
    pairing.testing.update_named_service("Light Strip", {CharacteristicsTypes.ON: True})

    assert callback.call_args_list == [mock.call({(accessory.aid, on_char.iid): {"value": 1}})]


async def test_update_aid_iid_events():
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")
    controller = FakeController()
    pairing = await controller.add_paired_device(accessories, "alias")

    callback = mock.Mock()
    await pairing.subscribe([(1, 8)])
    pairing.dispatcher_connect(callback)

    # Simulate that the state was changed on the device itself.
    pairing.testing.update_aid_iid([(1, 8, True)])

    assert callback.call_args_list == [mock.call({(1, 8): {"value": 1}})]


async def test_events_are_filtered():
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")
    controller = FakeController()
    pairing = await controller.add_paired_device(accessories, "alias")

    await pairing.subscribe([(1, 10)])
    pairing.dispatcher_connect(accessories.process_changes)

    # Simulate that the state was changed on the device itself.
    pairing.testing.update_named_service("Light Strip", {CharacteristicsTypes.ON: True})

    assert accessories.aid(1).characteristics.iid(8).value == 1


async def test_camera():
    accessories = Accessories.from_file("tests/fixtures/koogeek_ls1.json")
    controller = FakeController()
    pairing = await controller.add_paired_device(accessories, "alias")
    result = await pairing.image(1, 640, 480)
    assert len(result) > 0



from aiohomekit.crypto import hkdf_derive


def test_derive():
    material = hkdf_derive(b"1" * 32, b"Pair-Verify-Encrypt-Salt", b"Pair-Verify-Encrypt-Info")

    assert material == (
        b'\x8fC1v\xe3N\x8c\xa2\x9c\x94\xaaa\xce\xf5"\x94$7/xq\xbf\x8c;M\xe9\xe2\xa5N\xf9\xe5\x08'
    )



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from aiohomekit.crypto.chacha20poly1305 import (
    ChaCha20Poly1305Decryptor,
    ChaCha20Poly1305Encryptor,
    DecryptionError,
)


def test_example2_8_2():
    # Test aus 2.8.2
    plain_text = (
        b"Ladies and Gentlemen of the class of '99: If I could offer you only one tip for the future, "
        b"sunscreen would be it."
    )
    aad = 0x50515253C0C1C2C3C4C5C6C7.to_bytes(length=12, byteorder="big")
    key = 0x808182838485868788898A8B8C8D8E8F909192939495969798999A9B9C9D9E9F.to_bytes(
        length=32, byteorder="big"
    )
    iv = 0x4041424344454647.to_bytes(length=8, byteorder="big")
    fixed = 0x07000000.to_bytes(length=4, byteorder="big")
    r_ = (
        bytes(
            [
                0xD3,
                0x1A,
                0x8D,
                0x34,
                0x64,
                0x8E,
                0x60,
                0xDB,
                0x7B,
                0x86,
                0xAF,
                0xBC,
                0x53,
                0xEF,
                0x7E,
                0xC2,
                0xA4,
                0xAD,
                0xED,
                0x51,
                0x29,
                0x6E,
                0x08,
                0xFE,
                0xA9,
                0xE2,
                0xB5,
                0xA7,
                0x36,
                0xEE,
                0x62,
                0xD6,
                0x3D,
                0xBE,
                0xA4,
                0x5E,
                0x8C,
                0xA9,
                0x67,
                0x12,
                0x82,
                0xFA,
                0xFB,
                0x69,
                0xDA,
                0x92,
                0x72,
                0x8B,
                0x1A,
                0x71,
                0xDE,
                0x0A,
                0x9E,
                0x06,
                0x0B,
                0x29,
                0x05,
                0xD6,
                0xA5,
                0xB6,
                0x7E,
                0xCD,
                0x3B,
                0x36,
                0x92,
                0xDD,
                0xBD,
                0x7F,
                0x2D,
                0x77,
                0x8B,
                0x8C,
                0x98,
                0x03,
                0xAE,
                0xE3,
                0x28,
                0x09,
                0x1B,
                0x58,
                0xFA,
                0xB3,
                0x24,
                0xE4,
                0xFA,
                0xD6,
                0x75,
                0x94,
                0x55,
                0x85,
                0x80,
                0x8B,
                0x48,
                0x31,
                0xD7,
                0xBC,
                0x3F,
                0xF4,
                0xDE,
                0xF0,
                0x8E,
                0x4B,
                0x7A,
                0x9D,
                0xE5,
                0x76,
                0xD2,
                0x65,
                0x86,
                0xCE,
                0xC6,
                0x4B,
                0x61,
                0x16,
            ]
        ),
        bytes(
            [
                0x1A,
                0xE1,
                0x0B,
                0x59,
                0x4F,
                0x09,
                0xE2,
                0x6A,
                0x7E,
                0x90,
                0x2E,
                0xCB,
                0xD0,
                0x60,
                0x06,
                0x91,
            ]
        ),
    )
    nonce = fixed + iv
    r = ChaCha20Poly1305Encryptor(key).encrypt(aad, nonce, plain_text)
    assert r[:-16] == r_[0], "ciphertext"
    assert r[-16:] == r_[1], "tag"

    plain_text_ = ChaCha20Poly1305Decryptor(key).decrypt(aad, nonce, r)
    assert plain_text == plain_text_

    with pytest.raises(DecryptionError):
        ChaCha20Poly1305Decryptor(key).decrypt(aad, nonce, r + bytes([0, 1, 2, 3]))



from aiohomekit.enum import EnumWithDescription


class EnumTest(EnumWithDescription):
    RED = 1, "The colour is red"
    BLUE = 2, "This colour is blue"


def test_value_isnt_tuple():
    assert EnumTest.RED.value == 1


def test_casting():
    assert EnumTest(1) == EnumTest.RED


def test_has_description():
    assert EnumTest.RED.description == "The colour is red"
    assert EnumTest(1).description == "The colour is red"



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json

from aiohomekit.http.response import HttpResponse


def parse(data):
    response = HttpResponse()
    for i in range(0, len(data)):
        response.parse(data[i])
        if response.is_read_completely():
            break

    assert response.is_read_completely()

    return response


def test_example1():
    parts = [
        bytearray(
            b"HTTP/1.1 200 OK\r\nContent-Type: application/hap+json\r\n"
            b"Transfer-Encoding: chunked\r\nConnection: keep-alive\r\n\r\n"
        ),
        bytearray(
            b'5f\r\n{"characteristics":[{"aid":1,"iid":10,"value":35},'
            b'{"aid":1,"iid":13,"value":36.0999984741211}]}\r\n'
        ),
        bytearray(b"0\r\n\r\n"),
    ]
    res = parse(parts)
    assert res.code == 200
    assert res.get_http_name() == "HTTP"
    assert (
        res.body
        == b'{"characteristics":[{"aid":1,"iid":10,"value":35},{"aid":1,"iid":13,"value":36.0999984741211}]}'
    )
    json.loads(res.body.decode())


def test_example2():
    parts = [
        bytearray(
            b"EVENT/1.0 200 OK\r\nContent-Type: application/hap+json\r\nTransfer-Encoding: chunked\r\n\r\n"
        ),
        bytearray(
            b'5f\r\n{"characteristics":[{"aid":1,"iid":10,"value":35},'
            b'{"aid":1,"iid":13,"value":33.2000007629395}]}\r\n'
        ),
        bytearray(b"0\r\n\r\n"),
    ]
    res = parse(parts)
    assert res.code == 200
    assert res.get_http_name() == "EVENT"
    assert (
        res.body
        == b'{"characteristics":[{"aid":1,"iid":10,"value":35},{"aid":1,"iid":13,"value":33.2000007629395}]}'
    )
    j = json.loads(res.body.decode())
    assert j["characteristics"][1]["value"] == 33.2000007629395


def test_example3():
    parts = [
        bytearray(
            b"HTTP/1.1 200 OK\r\nServer: BaseHTTP/0.6 Python/3.5.3\r\nDate: Mon, 04 Jun 2018 20:06:06 GMT\r\n"
            b'Content-Type: application/hap+json\r\nContent-Length: 3740\r\n\r\n{"accessories": [{"services": '
            b'[{"characteristics": [{"maxLen": 64, "type": "00000014-0000-1000-8000-0026BB765291", "format": "bool", '
            b'"description": "Identify", "perms": ["pw"], "maxDataLen": 2097152, "iid": 3}, {"maxLen": 64, "type": '
            b'"00000020-0000-1000-8000-0026BB765291", "format": "string", "description": "Manufacturer", "perms": '
            b'["pr"], "maxDataLen": 2097152, "value": "lusiardi.de", "iid": 4}, {"maxLen": 64, "type": '
            b'"00000021-0000-1000-8000-0026BB765291", "format": "string", "description": "Model", "perms": ["pr"], '
            b'"maxDataLen": 2097152, "value": "Demoserver", "iid": 5}, {"maxLen": 64, "type": '
            b'"00000023-0000-1000-8000-0026BB765291", "format": "string", "description": "Name", "perms": ["pr"], '
            b'"maxDataLen": 2097152, "value": "Notifier", "iid": 6}, {"maxLen": 64, "type": '
            b'"00000030-0000-1000-8000-0026BB765291", "format": "string", "description": "Serial Number", "'
        ),
        bytearray(
            b'perms": ["pr"], "maxDataLen": 2097152, "value": "0001", "iid": 7}, {"maxLen": 64, "type": '
            b'"00000052-0000-1000-8000-0026BB765291", "format": "string", "description": "Firmware Revision",'
            b' "perms": ["pr"], "maxDataLen": 2097152, "value": "0.1", "iid": 8}], "type": '
            b'"0000003E-0000-1000-8000-0026BB765291", "iid": 2}, {"characteristics": [{"maxLen": 64, "type": '
            b'"00000025-0000-1000-8000-0026BB765291", "format": "bool", "description": '
            b'"Switch state (on/off)", "perms": ["pw", "pr", "ev"], "maxDataLen": 2097152, "value": false, '
            b'"iid": 10}, {"maxDataLen": 2097152, "minStep": 1, "description": "Brightness in percent", '
            b'"unit": "percentage", "minValue": 0, "perms": ["pr", "pw", "ev"], "maxValue": 100, "maxLen": '
            b'64, "type": "00000008-0000-1000-8000-0026BB765291", "format": "int", "value": 0, "iid": 11}, '
            b'{"maxDataLen": 2097152, "minStep": 1, "description": "Hue in arc degrees", "unit": '
            b'"arcdegrees", "minValue": 0, "perms": ["pr", "pw", "ev"], "maxValue": 360, "maxLen": 64, '
            b'"type": "00000013-0000-1000-8000-0026BB765291", "form'
        ),
        bytearray(
            b'at": "float", "value": 0, "iid": 12}, {"maxDataLen": 2097152, "minStep": 1, "description": '
            b'"Saturation in percent", "unit": "percentage", "minValue": 0, "perms": ["pr", "pw", "ev"], '
            b'"maxValue": 100, "maxLen": 64, "type": "0000002F-0000-1000-8000-0026BB765291", "format": '
            b'"float", "value": 0, "iid": 13}], "type": "00000043-0000-1000-8000-0026BB765291", "iid": 9}], '
            b'"aid": 1}, {"services": [{"characteristics": [{"maxLen": 64, "type": '
            b'"00000014-0000-1000-8000-0026BB765291", "format": "bool", "description": "Identify", "perms": '
            b'["pw"], "maxDataLen": 2097152, "iid": 16}, {"maxLen": 64, "type": '
            b'"00000020-0000-1000-8000-0026BB765291", "format": "string", "description": "Manufacturer", '
            b'"perms": ["pr"], "maxDataLen": 2097152, "value": "lusiardi.de", "iid": 17}, {"maxLen": 64, '
            b'"type": "00000021-0000-1000-8000-0026BB765291", "format": "string", "description": "Model", '
            b'"perms": ["pr"], "maxDataLen": 2097152, "value": "Demoserver", "iid": 18}, {"maxLen": 64, '
            b'"type": "00000023-0000-1000-8000-0026BB765291", "format": "string"'
        ),
        bytearray(
            b', "description": "Name", "perms": ["pr"], "maxDataLen": 2097152, "value": "Dummy", "iid": 19},'
            b' {"maxLen": 64, "type": "00000030-0000-1000-8000-0026BB765291", "format": "string", '
            b'"description": "Serial Number", "perms": ["pr"], "maxDataLen": 2097152, "value": "0001", '
            b'"iid": 20}, {"maxLen": 64, "type": "00000052-0000-1000-8000-0026BB765291", "format": "string",'
            b' "description": "Firmware Revision", "perms": ["pr"], "maxDataLen": 2097152, "value": "0.1", '
            b'"iid": 21}], "type": "0000003E-0000-1000-8000-0026BB765291", "iid": 15}, {"characteristics": '
            b'[{"perms": ["pw", "pr"], "maxLen": 64, "minValue": 2, "type": '
            b'"00000023-0000-1000-8000-0026BB765291", "format": "float", "description": "Test", "minStep":'
            b' 0.25, "maxDataLen": 2097152, "iid": 23}], "type": "00000040-0000-1000-8000-0026BB765291", '
            b'"iid": 22}], "aid": 14}]}'
        ),
    ]
    res = parse(parts)
    assert res.code == 200
    assert res.get_http_name() == "HTTP"
    json.loads(res.body.decode())


def test_example4():
    parts = [
        bytearray(
            b"HTTP/1.1 200 OK\r\nServer: BaseHTTP/0.6 Python/3.5.3\r\nDate: Mon, 04 Jun 2018 21:38:07 "
            b'GMT\r\nContent-Type: application/hap+json\r\nContent-Length: 3740\r\n\r\n{"accessories": '
            b'[{"aid": 1, "services": [{"type": "0000003E-0000-1000-8000-0026BB765291", '
            b'"characteristics": [{"format": "bool", "maxLen": 64, "iid": 3, "description": "Identify", '
            b'"perms": ["pw"], "type": "00000014-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}, '
            b'{"value": "lusiardi.de", "format": "string", "maxLen": 64, "iid": 4, "description": '
            b'"Manufacturer", "perms": ["pr"], "type": "00000020-0000-1000-8000-0026BB765291", '
            b'"maxDataLen": 2097152}, {"value": "Demoserver", "format": "string", "maxLen": 64, '
            b'"iid": 5, "description": "Model", "perms": ["pr"], "type": '
            b'"00000021-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}, {"value": "Notifier", '
            b'"format": "string", "maxLen": 64, "iid": 6, "description": "Name", "perms": ["pr"], '
            b'"type": "00000023-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}, {"value": "0001", '
            b'"format": "string", "maxLen": 64, "iid":'
        ),
        bytearray(
            b' 7, "description": "Serial Number", "perms": ["pr"], "type": '
            b'"00000030-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}, {"value": "0.1", "format": '
            b'"string", "maxLen": 64, "iid": 8, "description": "Firmware Revision", "perms": ["pr"], "type": '
            b'"00000052-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}], "iid": 2}, {"type": '
            b'"00000043-0000-1000-8000-0026BB765291", "characteristics": [{"value": false, "format": "bool", '
            b'"maxLen": 64, "iid": 10, "description": "Switch state (on/off)", "perms": ["pw", "pr", "ev"], '
            b'"type": "00000025-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}, {"maxValue": 100, '
            b'"format": "int", "minStep": 1, "description": "Brightness in percent", "perms": ["pr", "pw", '
            b'"ev"], "maxDataLen": 2097152, "type": "00000008-0000-1000-8000-0026BB765291", "maxLen": 64, '
            b'"iid": 11, "value": 0, "unit": "percentage", "minValue": 0}, {"maxValue": 360, "format": '
            b'"float", "minStep": 1, "description": "Hue in arc degrees", "perms": ["pr", "pw", "ev"], '
            b'"maxDataLen": 2097152, "type": "00000013-0000-1000'
        ),
        bytearray(
            b'-8000-0026BB765291", "maxLen": 64, "iid": 12, "value": 0, "unit": "arcdegrees", "minValue": 0},'
            b' {"maxValue": 100, "format": "float", "minStep": 1, "description": "Saturation in percent", '
            b'"perms": ["pr", "pw", "ev"], "maxDataLen": 2097152, "type": '
            b'"0000002F-0000-1000-8000-0026BB765291", "maxLen": 64, "iid": 13, "value": 0, "unit": '
            b'"percentage", "minValue": 0}], "iid": 9}]}, {"aid": 14, "services": [{"type": '
            b'"0000003E-0000-1000-8000-0026BB765291", "characteristics": [{"format": "bool", "maxLen": 64, '
            b'"iid": 16, "description": "Identify", "perms": ["pw"], "type": '
            b'"00000014-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}, {"value": "lusiardi.de", '
            b'"format": "string", "maxLen": 64, "iid": 17, "description": "Manufacturer", "perms": ["pr"], '
            b'"type": "00000020-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}, {"value": "Demoserver",'
            b' "format": "string", "maxLen": 64, "iid": 18, "description": "Model", "perms": ["pr"], "type": '
            b'"00000021-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}, {"value": "Dummy", "fo'
        ),
        bytearray(
            b'rmat": "string", "maxLen": 64, "iid": 19, "description": "Name", "perms": ["pr"], "type": '
            b'"00000023-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}, {"value": "0001", "format": '
            b'"string", "maxLen": 64, "iid": 20, "description": "Serial Number", "perms": ["pr"], "type": '
            b'"00000030-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}, {"value": "0.1", "format": '
            b'"string", "maxLen": 64, "iid": 21, "description": "Firmware Revision", "perms": ["pr"], '
            b'"type": "00000052-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}], "iid": 15}, {"type": '
            b'"00000040-0000-1000-8000-0026BB765291", "characteristics": [{"minStep": 0.25, "format": '
            b'"float", "maxLen": 64, "iid": 23, "description": "Test", "perms": ["pw", "pr"], "minValue": 2, '
            b'"type": "00000023-0000-1000-8000-0026BB765291", "maxDataLen": 2097152}], "iid": 22}]}]}'
        ),
    ]
    res = parse(parts)
    assert res.code == 200
    assert res.get_http_name() == "HTTP"
    json.loads(res.body.decode())


def test_example_lines():
    parts = [
        b"HTTP/1.1 200 OK\r\n",
        b"Content-Type: application/hap+json\r\n",
        b"Content-Length: 95\r\n",
        b"Connection: keep-alive\r\n",
        b"\r\n",
        b'{"characteristics":[{"aid":1,"iid":10,"value":35},',
        b'{"aid":1,"iid":13,"value":36.0999984741211}]}',
    ]
    res = parse(parts)
    assert res.code == 200
    assert res.get_http_name() == "HTTP"
    assert (
        res.body
        == b'{"characteristics":[{"aid":1,"iid":10,"value":35},{"aid":1,"iid":13,"value":36.0999984741211}]}'
    )
    json.loads(res.body.decode())


def test_example_204():
    res = parse([b"HTTP/1.1 204 No content\r\n\r\n"])
    assert res.code == 204
    assert res.get_http_name() == "HTTP"
    assert res.body == b""
    assert res.headers == []


def test_example_470():
    res = parse([b"HTTP/1.1 470 Connection Authorization Required\r\n\r\n"])
    assert res.code == 470
    assert res.get_http_name() == "HTTP"
    assert res.body == b""
    assert res.headers == []


def test_example_multi_status():
    res = parse(
        [
            (
                b"HTTP/1.1 207 Multi-Status\r\n"
                b"Content-Type: application/hap+json\r\n"
                b"Date: Sun, 29 Mar 2020 21:58:41 GMT\r\n"
                b"Connection: keep-alive\r\n"
                b"Transfer-Encoding: chunked\r\n"
                b"\r\n"
                b"33\r\n"
                b'{"characteristics":[{"aid":2,"iid":10,"status":0}]}\r\n'
                b"0\r\n"
                b"\r\n"
            )
        ]
    )
    assert res.code == 207
    assert res.body == b'{"characteristics":[{"aid":2,"iid":10,"status":0}]}'


def test_example_multi_status_2():
    res = parse(
        [
            (
                b"HTTP/1.1 207 Multi-Status\r\n"
                b"Content-Type: application/hap+json\r\n"
                b"Date: Sun, 29 Mar 2020 21:58:24 GMT\r\n"
                b"Connection: keep-alive\r\n"
                b"Transfer-Encoding: chunked\r\n"
                b"\r\n"
                b"40\r\n"
                b'{"characteristics":[{"aid":2,"iid":10,"value":true,"status":0}]}\r\n'
                b"0\r\n"
                b"\r\n"
            )
        ]
    )
    assert res.code == 207
    assert res.body == b'{"characteristics":[{"aid":2,"iid":10,"value":true,"status":0}]}'


def test_example_multi_status_3():
    """
    Test that 2 payloads arriving in same packet works.

    https://github.com/Jc2k/aiohomekit/issues/1
    """
    res = parse(
        [
            (
                b"HTTP/1.1 207 Multi-Status\r\n"
                b"Content-Type: application/hap+json\r\n"
                b"Date: Sun, 29 Mar 2020 21:57:54 GMT\r\n"
                b"Connection: keep-alive\r\n"
                b"Transfer-Encoding: chunked\r\n"
                b"\r\n"
                b"33\r\n"
                b'{"characteristics":[{"aid":2,"iid":10,"status":0}]}\r\n'
                b"0\r\n"
                b"\r\n"
                b"HTTP/1.1 207 Multi-Status\r\n"
                b"Content-Type: application/hap+json"
                b"\r\nDate: Sun, 29 Mar 2020 21:57:54 GMT\r\n"
                b"Connection: keep-alive\r\n"
                b"Transfer-Encoding: chunked\r\n"
                b"\r\n"
                b"33\r\n"
                b'{"characteristics":[{"aid":2,"iid":10,"status":0}]}\r\n'
                b"0\r\n"
                b"\r\n"
            )
        ]
    )
    assert res.code == 207


def test_newline_in_payload():
    parts = [
        b"HTTP/1.1 200 OK\r\n",
        b"Content-Type: application/hap+json\r\n",
        b"Content-Length: 97\r\n",
        b"Connection: keep-alive\r\n",
        b"\r\n",
        b'{"characteristics":[{"aid":1,"iid":10,"value":35},\r\n',
        b'{"aid":1,"iid":13,"value":36.0999984741211}]}',
    ]
    res = parse(parts)
    assert res.code == 200
    assert res.get_http_name() == "HTTP"
    assert (
        res.body
        == b'{"characteristics":[{"aid":1,"iid":10,"value":35},\r\n{"aid":1,"iid":13,"value":36.0999984741211}]}'
    )



from collections.abc import Iterable
import contextlib
import socket
from typing import Optional
from unittest.mock import patch

import pytest
from zeroconf import DNSQuestionType, Zeroconf
from zeroconf.asyncio import AsyncServiceInfo, AsyncZeroconf

from aiohomekit.characteristic_cache import CharacteristicCacheMemory
from aiohomekit.controller.ip.controller import IpController
from aiohomekit.exceptions import AccessoryNotFoundError
from aiohomekit.model.categories import Categories
from aiohomekit.model.status_flags import StatusFlags

HAP_TYPE_TCP = "_hap._tcp.local."
HAP_TYPE_UDP = "_hap._udp.local."
CLASS_IN = 1
TYPE_PTR = 12


class MockedAsyncServiceInfo(AsyncServiceInfo):
    async def async_request(
        self,
        zc: "Zeroconf",
        timeout: float,
        question_type: Optional[DNSQuestionType] = None,
    ) -> bool:
        return self.load_from_cache(zc)


def _get_mock_service_info():
    desc = {
        b"c#": b"1",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"s#": b"1",
        b"ci": b"5",
        b"sf": b"0",
    }
    return MockedAsyncServiceInfo(
        HAP_TYPE_TCP,
        f"foo.{HAP_TYPE_TCP}",
        addresses=[socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )


@contextlib.contextmanager
def _install_mock_service_info(
    mock_asynczeroconf: AsyncZeroconf, info: MockedAsyncServiceInfo
) -> Iterable[AsyncServiceInfo]:
    zeroconf: Zeroconf = mock_asynczeroconf.zeroconf
    zeroconf.cache.async_add_records(
        [*info.dns_addresses(), info.dns_pointer(), info.dns_service(), info.dns_text()]
    )

    assert zeroconf.cache.async_all_by_details(HAP_TYPE_TCP, TYPE_PTR, CLASS_IN) is not None

    with patch("aiohomekit.zeroconf.AsyncServiceInfo", MockedAsyncServiceInfo):
        yield


async def test_discover_find_one(mock_asynczeroconf: AsyncZeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)
    with _install_mock_service_info(mock_asynczeroconf, _get_mock_service_info()):
        async with controller:
            result = await controller.async_find("00:00:01:00:00:02")
        await controller._async_update_from_cache(mock_asynczeroconf.zeroconf)

    assert result.description.id == "00:00:01:00:00:02"
    assert result.description.category == Categories.LIGHTBULB
    assert result.description.config_num == 1
    assert result.description.state_num == 1
    assert result.description.model == "unittest"
    assert result.description.status_flags == StatusFlags(0)
    assert result.paired is True


async def test_async_reachable(mock_asynczeroconf: AsyncZeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)
    with _install_mock_service_info(mock_asynczeroconf, _get_mock_service_info()):
        async with controller:
            result = await controller.async_reachable("00:00:01:00:00:02")

    assert result is True


async def test_async_reachable_not_reachable(mock_asynczeroconf: AsyncZeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)
    with patch("aiohomekit.zeroconf.AsyncServiceInfo", MockedAsyncServiceInfo):
        async with controller:
            result = await controller.async_reachable("00:00:01:00:00:02", timeout=0.001)

    assert result is False


async def test_discover_find_one_unpaired(mock_asynczeroconf: AsyncZeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)

    svc = _get_mock_service_info()
    svc.properties[b"sf"] = b"1"
    svc._set_properties(svc.properties)
    with _install_mock_service_info(mock_asynczeroconf, svc):
        async with controller:
            result = await controller.async_find("00:00:01:00:00:02")
        await controller._async_update_from_cache(mock_asynczeroconf.zeroconf)

    assert result.description.id == "00:00:01:00:00:02"
    assert result.description.status_flags == StatusFlags.UNPAIRED
    assert result.paired is False


async def test_discover_find_none(mock_asynczeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)

    async with controller:
        with pytest.raises(AccessoryNotFoundError):
            await controller.async_find("00:00:00:00:00:00", timeout=0.001)


async def test_find_device_id_case_lower(mock_asynczeroconf: AsyncZeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)

    svc_info_1 = _get_mock_service_info()
    svc_info_1.properties[b"id"] = b"aa:aa:aa:aa:aa:aa"
    svc_info_1._set_properties(svc_info_1.properties)

    with _install_mock_service_info(mock_asynczeroconf, svc_info_1):
        async with controller:
            await controller._async_update_from_cache(mock_asynczeroconf.zeroconf)
            res = await controller.async_find("AA:AA:AA:AA:AA:AA")
            assert res.description.id == "aa:aa:aa:aa:aa:aa"

    svc_info_2 = _get_mock_service_info()
    svc_info_2.properties[b"id"] = b"aa:aa:aa:aa:aa:aa"
    svc_info_2._set_properties(svc_info_2.properties)

    with _install_mock_service_info(mock_asynczeroconf, svc_info_2):
        svc_info_2.properties[b"id"] = b"aa:aa:aa:aa:aa:aa"

        async with controller:
            await controller._async_update_from_cache(mock_asynczeroconf.zeroconf)
            res = await controller.async_find("aa:aa:aa:aa:aa:aa")
            assert res.description.id == "aa:aa:aa:aa:aa:aa"


async def test_find_device_id_case_upper(mock_asynczeroconf: AsyncZeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)

    svc_info = _get_mock_service_info()
    svc_info.properties[b"id"] = b"AA:AA:aa:aa:AA:AA"
    svc_info._set_properties(svc_info.properties)

    with _install_mock_service_info(mock_asynczeroconf, svc_info):
        async with controller:
            await controller._async_update_from_cache(mock_asynczeroconf.zeroconf)
            res = await controller.async_find("AA:AA:AA:AA:AA:AA")
            assert res.description.id == "aa:aa:aa:aa:aa:aa"

    svc_info = _get_mock_service_info()
    svc_info.properties[b"id"] = b"AA:AA:aa:aa:AA:AA"
    svc_info._set_properties(svc_info.properties)

    with _install_mock_service_info(mock_asynczeroconf, svc_info):
        async with controller:
            await controller._async_update_from_cache(mock_asynczeroconf.zeroconf)
            res = await controller.async_find("aa:aa:aa:aa:aa:aa")
            assert res.description.id == "aa:aa:aa:aa:aa:aa"


async def test_discover_discover_one(mock_asynczeroconf: AsyncZeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)

    srv_info = _get_mock_service_info()
    with _install_mock_service_info(mock_asynczeroconf, srv_info):
        async with controller:
            await controller._async_update_from_cache(mock_asynczeroconf.zeroconf)
            results = [d async for d in controller.async_discover()]

    assert results[0].description.id == "00:00:01:00:00:02"
    assert results[0].description.category == Categories.LIGHTBULB
    assert results[0].description.config_num == 1
    assert results[0].description.state_num == 1
    assert results[0].description.model == "unittest"
    assert results[0].description.status_flags == StatusFlags(0)
    assert results[0].paired is True


async def test_discover_none(mock_asynczeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)

    results = [d async for d in controller.async_discover()]
    assert results == []


async def test_discover_missing_csharp(mock_asynczeroconf: AsyncZeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)

    svc_info = _get_mock_service_info()
    del svc_info.properties[b"c#"]
    svc_info._set_properties(svc_info.properties)

    with _install_mock_service_info(mock_asynczeroconf, svc_info):
        async with controller:
            await controller._async_update_from_cache(mock_asynczeroconf.zeroconf)
            results = [d async for d in controller.async_discover()]

    assert results[0].description.id == "00:00:01:00:00:02"
    assert results[0].description.config_num == 0


async def test_discover_csharp_case(mock_asynczeroconf: AsyncZeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)

    svc_info = _get_mock_service_info()
    del svc_info.properties[b"c#"]
    svc_info.properties[b"C#"] = b"1"
    svc_info._set_properties(svc_info.properties)

    with _install_mock_service_info(mock_asynczeroconf, svc_info):
        async with controller:
            await controller._async_update_from_cache(mock_asynczeroconf.zeroconf)
            results = [d async for d in controller.async_discover()]

    assert results[0].description.config_num == 1


async def test_discover_device_id_case_lower(mock_asynczeroconf: AsyncZeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)

    svc_info = _get_mock_service_info()
    svc_info.properties[b"id"] = b"aa:aa:aa:aa:aa:aa"
    svc_info._set_properties(svc_info.properties)

    with _install_mock_service_info(mock_asynczeroconf, svc_info):
        async with controller:
            await controller._async_update_from_cache(mock_asynczeroconf.zeroconf)

            results = [d async for d in controller.async_discover()]

    assert results[0].description.id == "aa:aa:aa:aa:aa:aa"


async def test_discover_device_id_case_upper(mock_asynczeroconf: AsyncZeroconf):
    controller = IpController(char_cache=CharacteristicCacheMemory(), zeroconf_instance=mock_asynczeroconf)

    svc_info = _get_mock_service_info()
    svc_info.properties[b"id"] = b"AA:AA:aa:aa:AA:AA"
    svc_info._set_properties(svc_info.properties)

    with _install_mock_service_info(mock_asynczeroconf, svc_info):
        async with controller:
            await controller._async_update_from_cache(mock_asynczeroconf.zeroconf)

            results = [d async for d in controller.async_discover()]

    assert results[0].description.id == "aa:aa:aa:aa:aa:aa"
